/*****************************************************************************
 * Copyright (C) 1997-2009, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   array.h
 *		- array info routines
 *
 ******************************************************************************
 */

#ifndef ARRAY_H
#define ARRAY_H

#include "cnode_def.h"
#include "cdecl.h"
#include "cgenvar.h"

/**
 * Class method to query array info.
 */
class Array {
public:
	/**
 	 * Get width of array.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return width of expression
 	 */
	static int Width( CNode* ref, CNode* index )
	{ 
	    CNode* select = GetSelect( ref, index );
	    if( select ) {
		switch( select->GetOp() ) {
		case eSLICE:
		case ePSLICE:
		case eMSLICE:
		    return select->GetWidth();
		default:
		    return 1;
		}
	    }
	    return ref->GetWidth();
	}
	/**
 	 * Get width of array as an expression.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return width expression.
 	 */
	static CNode* WidthExp( CNode* ref, CNode* index )
	{ 
	    CNode* select = GetSelect( ref, index );
	    if( select ) {
		switch( select->GetOp() ) {
		case eSLICE:
		case ePSLICE:
		case eMSLICE:
		    return select->GetWidthExp();
		default:
		    return cINT32(1);
		}
	    }
	    return ref->GetWidthExp();
	}
	/**
 	 * Determine if width of array variable is constant.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width is constant. ie dependent upon only
 	 *         on parameters and constants.
 	 */
	static int WidthConstant( CNode* ref, CNode* index ) 
	{ 
	    CNode* select = GetSelect( ref, index );
	    if( select ) {
		switch( select->GetOp() ) {
		case eSLICE:
		case ePSLICE:
		case eMSLICE:
		    return select->IsWidthConstant();
		default:
		    return TRUE;
		}
	    }
	    return ref->IsWidthConstant();
	}
	/**
 	 * Determine if width of array variable is volatile.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width is volatile. ie dependent upon only
 	 *         on parameters or variables.
 	 */
	static int WidthVolatile( CNode* ref, CNode* index ) 
	{ 
	    CNode* select = GetSelect( ref, index );
	    if( select ) {
		switch( select->GetOp() ) {
		case eSLICE:
		case ePSLICE:
		case eMSLICE:
		    return select->IsWidthVolatile();
		default:
		    return FALSE;
		}
	    }
	    return ref->IsWidthVolatile();
	}
	/**
 	 * Determine if width of array can be evaluated.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width can be evaluated.
 	 */
	static int WidthEvaluateable( CNode* ref, CNode* index ) 
	{ 
	    CNode* select = GetSelect( ref, index );
	    if( select ) {
		switch( select->GetOp() ) {
		case eSLICE:
		case ePSLICE:
		case eMSLICE:
		    return select->IsWidthEvaluateable();
		default:
		    return TRUE;
		}
	    }
	    return ref->IsWidthEvaluateable();
	}
	/**
 	 * Determine type of array.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return type of array return value.
 	 */
	static NodeType_t Type( CNode* ref, CNode* index )
	{ 
	    CNode* select = GetSelect( ref, index );
	    if( select ) {
	        return eB;
	    }
	    return ref->GetNodeType();
	}
	/**
 	 * Get decl for array reference
 	 * \param ref decl reference node
 	 * \return decl associated with the reference
 	 */
	static CDecl* ArrayDecl( CNode* ref )
	{
	    MASSERT( ref );
	    switch( ref->GetOp() ) {
	    case eNET_REF:		
		return ref->Arg<CNet*>(0);
	    case eVAR_REF:		
		return ref->Arg<CVar*>(0);
	    case ePARAM_REF:		
		return ref->Arg<CParam*>(0);
	    case ePORT_REF:		
		return ref->Arg<CPortDir*>(0);
	    case eFWD_REF:		
		return ref->Arg<CFref*>(0);
	    case eGENVAR_REF:		
		return ref->Arg<CGenvar*>(0);
	    default:
		MASSERT( FALSE );
	    }
	}
	/**
 	 * Get array select (if present).
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return select expression, NULL if not present
 	 */
	static CNode* GetSelect( CNode* ref, CNode* index )
	{ 
	    vector<CNode*> indexes;
	    EList2VectorExclude( index, set<NodeOp_t>(), indexes );
	    CNode* select = indexes.back();
	    while( select && select->GetOp() == eMACRO_EXPR ) {
		select = select->Arg<CNode*>(1);
	    }
	    if( !select ) {
		select = indexes.back();
	    }
	    switch( select->GetOp() ) {
	    case eSLICE:
	    case ePSLICE:
	    case eMSLICE:
		return select;
	    default:
		break;
	    }
	    if( ref->GetOp() != eEXTERNAL_REF && 
	 	ListCount(index) > ArrayDecl(ref)->
					GetNumberOfUnpackedDimensions() ) {
		return select;
	   }
	   return NULL;
	}
};

#endif // ARRAY_H
