/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdatatype.h
 *		- abstract class for data types 
 *
 *
 ******************************************************************************
 */

#ifndef CDATATYPE_H
#define CDATATYPE_H

#include "cobject.h"

class CNode;

/**
 * Variable data types
 */
enum VarDataType_t {
	eNO_VAR_TYPE = 0,///< no type declaration
	eREAL,		 ///< real declaration
	eSHORTREAL,	 ///< shortreal declaration
	eREALTIME,	 ///< realtime declaration
	eREG,	         ///< register declaration
	eTIME,		 ///< time declaration
	eLOGIC,		 ///< logic declaration
	eINTEGER,	 ///< integer declaration
	eBIT,		 ///< integer declaration
	eINT,		 ///< int declaration
	eBYTE,		 ///< byte declaration
	eSHORTINT,	 ///< shortint declaration
	eLONGINT,	 ///< longint declaration
	eDEVENT,	 ///< event declaration
};
#if defined CDATATYPE_CC
/**
 * Array to convert DataType_t to character string
 */
const char* varDataTypeName[] = {
	"",
	"real",
	"shortreal",
	"realtime",
	"reg",
	"time",
	"logic",
	"integer",
	"bit",
	"int",
	"byte",
	"shortint",
	"longint",
	"event",
};
#else // CDATATYPE_CC
extern const char* varDataTypeName[];
#endif // CDATATYPE_CC

/**
 * Expression node type
 */
enum NodeType_t {
	eU,	///< undefined
	eR,	///< real - have width 0
	eS,	///< signed bit vector, includes integer
	eE,	///< event - have width 0
	eB	///< unsigned bit vector
};

/**
 * Base class for describing data types
 */
class CDataType : public CObject
{
public:
private:
	int	      const_;	  ///< declaration is a const type
	VarDataType_t varDataType;  ///< variable data type
public:
        /**
         * Create data type instance
         */
        CDataType( VarDataType_t t );
	/**
 	 * Create a new copy with a deep copy. 
 	 * \param heap heap to use for allocation.
 	 */
	virtual CDataType* Clone( CObstack* heap ) = 0;
	/**
 	 * Set declaration's variable data type
 	 * \param v variable data type
 	 */
	virtual void SetVarDataType( VarDataType_t v ) { varDataType = v; }
	/**
 	 * Get declaration's variable data type
 	 * \return value of signed property.
 	 */
	virtual VarDataType_t  GetVarDataType() const { return varDataType; }
	/**
 	 * Set declaration's const property.
 	 * Use to indicate type is a constant variable.
 	 * \param v value to const property.
 	 */
	virtual void SetConst( int v ) { const_ = v; }
	/**
 	 * Get declaration's const property.
 	 * \return value of const property.
 	 */
	virtual int  GetConst() const { return const_; }
	/**
 	 * Get declaration's 2 state property.
 	 * \return value of 2 state property.
 	 */
	virtual int  GetTwoState() const = 0;
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() const = 0;
        /**
         * Get data type
         * \return node type
         */
        virtual NodeType_t GetNodeType( void ) const = 0;
        /**
         * Determine if complete data structure is packed
         * \return non-zero if complete data structure is packed.
         */
        virtual int IsPacked() const = 0;
        /**
         * Get expression for datatype's overall packed range (msb/lsb)
         * \return range expression, NULL if scalar.
         */
        virtual CNode* GetPackedRange() const = 0;
        /**
         * Dump data type info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) const = 0;
        /**
         * Walk tree invoking callback on each node before children have 
         * been visited. callback can terminate transversal early.
         * callback returns 0 to terminate transversal early.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PreVisit1( int (*callback)(CNode*,void*), void* data ) =0;
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PostVisit1( void (*callback)(CNode*,void*), void* data ) = 0;
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited. callback may modify tree and return a modified
         * subtree.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         * \return modified child subtree.
         */
	virtual void PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data ) = 0;
protected:
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CDataType& o );
private:
	/*
 	 * disable copy constructors
 	 */
	CDataType( const CDataType& o );

};

#endif // CDATATYPE_H

