/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdecl.cpp
 *		- methods for declarations 
 *
 *
 ******************************************************************************
 */

#define CDECL_CPP

#include "cdecl.h"
#include "cnode.h"
#include "cattr.h"
#include "cdatatype.h"

/************************************************************
  	Copy
	- perform deep copy from given object to this object
 ***********************************************************/

void CDecl::Copy( CObstack* heap, const CDecl& o )
{
    loc           = o.loc;
    symbol        = o.symbol;
    type          = o.type;
    declStatement = o.declStatement;
    attributes    = o.attributes ? o.attributes->Clone(heap) : NULL;
    pragmas       = o.pragmas ? o.pragmas->Clone(heap) : NULL;
    numberOfUnpackedDimensions = o.numberOfUnpackedDimensions;
    for( int i = 0; i < numberOfUnpackedDimensions; ++i ) {
        unpackedRange[i] = o.unpackedRange[i] ? o.unpackedRange[i]->Clone(heap) : NULL;
    }
    flags = o.flags;
    automatic = o.automatic;
    constAttr = o.constAttr;
    wireAttr = o.wireAttr;
    varAttr = o.varAttr;
    dataType = o.dataType ? o.dataType->Clone(heap) : NULL;
}

/***********************************************************
	IsWidthConstant
	- return TRUE if width is a constant expression
 **********************************************************/

int  CDecl::IsWidthConstant( void ) {
	CNode* lsb = GetLsb();
	CNode* msb = GetMsb();

	if( msb == NULL ) {
		MASSERT( lsb == NULL );
		return TRUE;
	}
		
	return msb->IsWidthConstant() && lsb->IsWidthConstant(); 
}
/***********************************************************
	IsWidthVolatile
	- return TRUE if width is a volatile expression
 **********************************************************/

int  CDecl::IsWidthVolatile( void ) {
	CNode* lsb = GetLsb();
	CNode* msb = GetMsb();

	if( msb == NULL ) {
		MASSERT( lsb == NULL );
		return FALSE;
	}
		
	return msb->IsWidthVolatile() || lsb->IsWidthVolatile(); 
}
/***********************************************************
	IsWidthEvaluateable
	- return TRUE if width is an evaluateable expression
 **********************************************************/

int  CDecl::IsWidthEvaluateable( void ) {
	CNode* lsb = GetLsb();
	CNode* msb = GetMsb();

	if( msb == NULL ) {
		MASSERT( lsb == NULL );
		return TRUE;
	}
		
	return msb->IsEvaluateable() && lsb->IsEvaluateable(); 
}
/***********************************************************
	GetWidth
	- evaluate current decl width. Expression
	  must be constant.  
 **********************************************************/

INT32 CDecl::GetWidth( void ) {
	CNode* lsb = GetLsb();
	CNode* msb = GetMsb();
	
	if( msb == NULL ) {
		MASSERT( lsb == NULL );
		return 1;
	} 
	if( msb == lsb ) {	
		return 1;
	}
	INT32 l = GetLsbInt();
	INT32 m = GetMsbInt();
	return m>=l ? m-l+1 : l-m+1;
}
/***********************************************************
	GetWidthExp
	- return expression for width 
 **********************************************************/

CNode* CDecl::GetWidthExp( void ) {
	CNode* lsb = GetLsb();
	CNode* msb = GetMsb();
	
	if( msb == NULL ||
	    lsb == msb ) {
		return cINT32(1);
	} 
	MASSERT(lsb);
	MASSERT(msb);
	if( !lsb->IsVolatile() && !msb->IsVolatile() &&
	    lsb->IsEvaluateable() && msb->IsEvaluateable() ) {
	    return cINT32(GetWidth());
	}
        return cADD(cABS(cSUB(msb->Clone(),lsb->Clone())),cINT32(1));
}
/***********************************************************
	WidthDirection
	- evaluate current decl width direction.
	  Expression must be constant.  
	  1 - big endian, -1 - little endian, 0 - cannot be determined
 **********************************************************/

int CDecl::WidthDirection( void ) {
	CNode* lsb = GetLsb();
	CNode* msb = GetMsb();
	
	if( msb == NULL ) {
		MASSERT( lsb == NULL );
		return 1;
	} 
	if( msb == lsb ) {	
		return 1;
	}
	if( !GetLsb()->IsEvaluateable() ||
	    !GetMsb()->IsEvaluateable() ) {
	    return 0;
	}
	INT32 l = GetLsbInt();
	INT32 m = GetMsbInt();
	return m>=l ? 1 : -1;
}
/***********************************************************
	GetUnpackedMsi
 	- Get expression tree for upper limit of unpacked array dimension.
 **********************************************************/
CNode* CDecl::GetUnpackedMsi( INT32 dim ) 
{ 
   return unpackedRange[dim] ? 
		(CNode*)unpackedRange[dim]->Arg<CNode*>(0) : NULL; 
}
/***********************************************************
	GetUnpackedLsi
 	- Get expression tree for lower limit of unpacked array dimension.
 **********************************************************/
CNode* CDecl::GetUnpackedLsi( INT32 dim ) 
{ 
   return unpackedRange[dim] ? 
		(CNode*)unpackedRange[dim]->Arg<CNode*>(1) : NULL; 
}
/***********************************************************
	GetMsb
	- Get declarations msb specification 
	  NULL for scalars. 
 **********************************************************/
CNode* CDecl::GetMsb() 
{ 
    CNode* range = dataType ? dataType->GetPackedRange() : NULL;
    if( !range ) {
	return NULL;
    }
    if( range->GetOp() == eMACRO_EXPR ) {
	return range->Arg<CNode*>(1)->Arg<CNode*>(0);
    }
    return range->Arg<CNode*>(0);
}
/***********************************************************
	GetLsb
	- Get declarations lsb specification 
	  NULL for scalars. 
 **********************************************************/
CNode* CDecl::GetLsb() 
{ 
    CNode* range = dataType ? dataType->GetPackedRange() : NULL;
    if( !range ) {
	return NULL;
    }
    if( range->GetOp() == eMACRO_EXPR ) {
	return range->Arg<CNode*>(1)->Arg<CNode*>(1);
    }
    return range->Arg<CNode*>(1);
}
/***********************************************************
	GetMsbInt
	- evaluate current decl msb. Expression
	  must be constant.  
 **********************************************************/

INT32 CDecl::GetMsbInt( void ) {
	CNode* exp;
	CVector width(32);

	width.Signed( TRUE );
	CNode::UseEvalStack();
	exp = GetMsb()->Clone();
	MASSERT( exp->IsConstant() );
	exp->EvalVector( width );
	CNode::ResetBuildStack();
	return width.GetINT32();
}

/***********************************************************
	GetLsbInt
	- evaluate current decl lsb. Expression
	  must be constant.  
 **********************************************************/

INT32 CDecl::GetLsbInt( void ) {
	CNode* exp;
	CVector width(32);

	width.Signed( TRUE );
	CNode::UseEvalStack();
	exp = GetLsb()->Clone();
	MASSERT( exp->IsConstant() );
	exp->EvalVector( width );
	CNode::ResetBuildStack();
	return width.GetINT32();
}

/***********************************************************
	HasAttribute
	- returns 1 if attribute is present
 **********************************************************/
int CDecl::HasAttribute( char* name, CNode* n, int init )
{
    int result = 0;
    if( !name ) {
	return 0;
    }
    if( init ) {
	n = GetAttributes();
    }
    if( !n ) {
	return 0;
    }
    switch( n->GetOp() ) {
    case eLIST:
	result |= HasAttribute( name, n->Arg<CNode*>(0), 0 );
	result |= HasAttribute( name, n->Arg<CNode*>(1), 0 );
	break;
    case eATTRIBUTE:
	{
	    CAttr* attr = n->Arg<CAttr*>(0);
	    return attr->HasAttribute( name );
	}
    default:
	return 0;
    }
   
    return result; 
}

/****************************************************
	GetNodeType
 	 - Get node type
*****************************************************/
NodeType_t CDecl::GetNodeType( void )
{
    return GetDataType()->GetNodeType();
}

/****************************************************
    PreVisit1
	- helper routine to support PreVisit1 node 
	  traversal.
*****************************************************/
void CDecl::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    for( int i = 0; i < GetNumberOfUnpackedDimensions(); ++i ) {
	GetUnpackedRange( i )->PreVisit1( func, data );
    }
    if( pragmas ) {
	pragmas->PreVisit1( func, data );
    }
    if( attributes ) {
	attributes->PreVisit1( func, data );
    }
    if( dataType ) {
	dataType->PreVisit1( func, data );
    }
}

/****************************************************
    PostVisit1
	- helper routine to support PostVisit1 node 
	  traversal.
*****************************************************/
void CDecl::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    for( int i = 0; i < GetNumberOfUnpackedDimensions(); ++i ) {
	GetUnpackedRange( i )->PostVisit1( func, data );
    }
    if( pragmas ) {
	pragmas->PostVisit1( func, data );
    }
    if( attributes ) {
	attributes->PostVisit1( func, data );
    }
    if( dataType ) {
	dataType->PostVisit1( func, data );
    }
}

/****************************************************
    PostSubVisit1
	- helper routine to support PostSubVisit1 node 
	  traversal.
*****************************************************/
void CDecl::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    for( int i = 0; i < GetNumberOfUnpackedDimensions(); ++i ) {
	SetUnpackedRange( i, GetUnpackedRange( i )->PostSubVisit1( func, data ) );
    }
    if( pragmas ) {
	pragmas = pragmas->PostSubVisit1( func, data );
    }
    if( attributes ) {
	attributes = attributes->PostSubVisit1( func, data );
    }
    if( dataType ) {
	dataType->PostSubVisit1( func, data );
    }
}

