/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdatatype.cc
 *		- abstract class for data types 
 *
 *
 ******************************************************************************
 */

#define CSIMPLETYPE_CC

#include "csimpletype.h"
#include "cnode.h"

/**
 * Create data type instance
 * This should never be call directly.
 */
CSimpleType::CSimpleType( VarDataType_t vt, CNode* aRange )
    : CDataType(vt),
      signedAttr(FALSE),
      unsignedAttr(FALSE),
      range(aRange)
{
   /*
    * ensure ranges that are specified with macros used defined macros
    */
   MASSERT( !range || range->GetOp() != eMACRO_EXPR || range->Arg<CNode*>(1) );
}

/**
 * Create a new copy with a deep copy. 
 */
CDataType* CSimpleType::Clone( CObstack* heap )
{
    CSimpleType* clone = new(heap) CSimpleType(GetVarDataType(), NULL);
    clone->Copy( heap, *this ); 
    return clone;
}

/**
 * Perform deep copy of given object to this one
 * This should never be call directly, only by subclasses.
 * \param o declaration to copy.
 */
void CSimpleType::Copy( CObstack* heap, const CSimpleType& o )
{
    CDataType::Copy( heap, o );
    signedAttr   = o.signedAttr;
    unsignedAttr = o.unsignedAttr;
    range    = o.range ? o.range->Clone(heap) : NULL;
}

/**
 * Get declaration's 2 state property.
 * \return value of 2 state property.
 */
int  CSimpleType::GetTwoState() const
{
    switch( GetVarDataType() ) {
    case eNO_VAR_TYPE:
    case eREG:
    case eTIME:
    case eLOGIC:          
    case eINTEGER:        
	return FALSE;
    case eDEVENT:
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
    case eINT:
    case eBIT:            
    case eBYTE:           
    case eSHORTINT:       
    case eLONGINT:      
	return TRUE;
    default:
	MASSERT( FALSE );
	return FALSE;
    }
}
/**
 * Get declartion's signed property.
 * \return value of signed property.
 */
int  CSimpleType::GetSigned() const
{
    if( signedAttr ) {
	return TRUE;
    }
    if( unsignedAttr ) {
	return FALSE;
    }

    switch( GetVarDataType() ) {
    case eNO_VAR_TYPE:
    case eREG:
    case eTIME:
    case eLOGIC:          
    case eDEVENT:
    case eBIT:            
	return FALSE;
    case eINTEGER:        
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
    case eINT:
    case eBYTE:           
    case eSHORTINT:       
    case eLONGINT:      
	return TRUE;
    default:
	MASSERT( FALSE );
	return FALSE;
    }
}

/**
 * Get data type
 * \return node type
 */
NodeType_t CSimpleType::GetNodeType( void ) const
{
    switch( GetVarDataType() ) {
    case eNO_VAR_TYPE:
    case eREG:
    case eTIME:
    case eINT:
    case eLOGIC:          
    case eINTEGER:        
    case eBIT:            
    case eBYTE:           
    case eSHORTINT:       
    case eLONGINT:      
	return GetSigned() ? eS : eB;
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
	return eR;
    case eDEVENT:
	return eU;
    default:
	MASSERT( FALSE );
	return eU;
    }
}

/**
 * Dump data type info to file descriptor.
 * \param f file descriptor.
 */
void CSimpleType::Dump( FILE* f ) const
{ 
    int dumpRange = FALSE;

    switch( GetVarDataType() ) {
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
    case eINT:
    case eINTEGER:
    case eBYTE:
    case eSHORTINT:
    case eLONGINT:
    case eTIME:
        break;
    case eBIT:
    case eNO_VAR_TYPE:
    case eREG:
    case eLOGIC:
        dumpRange = TRUE;
        break;
    case eDEVENT:
	dumpRange = FALSE;
    default:
        MASSERT( FALSE );
    }
    fprintf( f, "%s ", varDataTypeName[GetVarDataType()] );
    if( GetSignedAttr() ) {
        fprintf( f, "signed " );
    }
    if( GetUnsignedAttr() ) {
        fprintf( f, "unsigned " );
    }

    if( dumpRange ) {
        if( GetPackedRange() ) {
	    GetPackedRange()->Dump( f );
            fprintf( f, " " );
        }
   }

}

/**
 * Walk tree invoking callback on each node before children have 
 * been visited. callback can terminate transversal early.
 * callback returns 0 to terminate transversal early.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CSimpleType::PreVisit1( int (*callback)(CNode*,void*), void* data )
{
    if( range ) {
	range->PreVisit1( callback, data ); 
    }
}

/**
 * Walk tree invoking callback on each node after children have 
 * been visited.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CSimpleType::PostVisit1( void (*callback)(CNode*,void*), void* data )
{
    if( range ) {
	range->PostVisit1( callback, data ); 
    }
}

/**
 * Walk tree invoking callback on each node after children have 
 * been visited. callback may modify tree and return a modified
 * subtree.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 * \return modified child subtree.
 */
void CSimpleType::PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data )
{
    if( range ) {
	range = range->PostSubVisit1( callback, data ); 
    }
}

