/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   csimpletype.h
 *		- abstract class for simple built in data types 
 *
 *
 ******************************************************************************
 */

#ifndef CSIMPLETYPE_H
#define CSIMPLETYPE_H

#include "cobject.h"
#include "cdatatype.h"


/**
 * class for describing simple builtin data types
 */
class CSimpleType : public CDataType
{
public:
private:
	int        signedAttr;	  ///< TRUE if signed attribute is present
	int        unsignedAttr;  ///< TRUE if unsigned attribute is present
	CNode*	   range;	  ///< declarations packed range
public:
        /**
         * Create data type instance
         */
        CSimpleType( VarDataType_t vt, CNode* aRange );
	/**
 	 * Create a new copy with a deep copy. 
 	 * \param heap heap to use for allocation.
 	 */
	virtual CDataType* Clone( CObstack* heap );
	/**
 	 * Get declaration's signed attribute.
 	 * \return value of signed attribute.
 	 */
	virtual int GetSignedAttr() const { return signedAttr; }
	/**
 	 * Get declaration's unsigned attribute.
 	 * \return value of unsigned attribute.
 	 */
	virtual int GetUnsignedAttr() const { return unsignedAttr; }
	/**
 	 * Set declaration's signed attribute.
 	 * \param v value of signed attribute.
 	 */
	virtual void SetSignedAttr( int v ) { signedAttr = v; }
	/**
 	 * Set declaration's unsigned attribute.
 	 * \param v value of unsigned attribute.
 	 */
	virtual void SetUnsignedAttr( int v ) { unsignedAttr = v; }
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() const;
	/**
 	 * Get declaration's 2 state property.
 	 * \return value of 2 state property.
 	 */
	virtual int  GetTwoState() const;
        /**
         * Get data type
         * \return node type
         */
        virtual NodeType_t GetNodeType( void ) const;
        /**
         * Set expression for datatype's overall packed range (msb/lsb)
         * \param r expression to assign to datatype, NULL if scalar.
         */
        virtual void SetPackedRange( CNode* r ) { range = r; }
        /**
         * Determine if complete data structure is packed
         * \return non-zero if complete data structure is packed.
         */
        virtual int IsPacked() const { return TRUE; } 
        /**
         * Get expression for datatype's overall packed range (msb/lsb)
         * \return range expression, NULL if scalar.
         */
        virtual CNode* GetPackedRange() const { return range; }
        /**
         * Dump data type info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) const;
        /**
         * Walk tree invoking callback on each node before children have 
         * been visited. callback can terminate transversal early.
         * callback returns 0 to terminate transversal early.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PreVisit1( int (*callback)(CNode*,void*), void* data );
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PostVisit1( void (*callback)(CNode*,void*), void* data );
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited. callback may modify tree and return a modified
         * subtree.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         * \return modified child subtree.
         */
	virtual void PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data );
protected:
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CSimpleType& o );
private:
 	/*
 	 * disable copy constructor
 	 */
	CSimpleType( const CSimpleType& o ); 

};

#endif // CSIMPLETYPE_H

