/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cxprop.cpp
 *		- methods for verilog design rule checker 
 *
 ******************************************************************************
 */

/**
 *
 * \page xprop_plugin Xprop Tool
 * The xprop tool is specified using the -tool xprop switch on the 
 * command line. This filter will cause the verilog code (ASTs) to be 
 * instrumented such that X's are properly propagated by verilog 
 * simulation tools. This is used to reduce the simlulated differences 
 * between gatelevels representations and RTL. I a properly constructed 
 * design flow this greatly reduces the need for gate level simulation 
 * (which is much more expensive than RTL simulation). For a detailed 
 * description of the transformations done by vrq and the reasoning 
 * behind it see the \ref xpropRational page.
 *
 * \section xprop_switches Xprop Switches
 * Vrq adds instrumentation all constructs except clock gating. The 
 * switches below may be used to alter the default behavior:
 * - \b +xprop-disable-if Disable if instrumentation
 * - \b +xprop-disable-case Disable case instrumentation
 * - \b +xprop-disable-select Disable bit/part select instrumentation
 * - \b +xprop-disable-array Disable array assignment instrumentation
 * - \b +xprop-disable-ternary Disable ’?’ instrumentation
 * - \b +xprop-enable-clock Enable instrumentation of clocks 
 *   (to check clock gaters)
 *
 * The following switches alter how variables and constructs are treated. 
 * By default all variables are assume to be able to be 0,1,x,z. You can 
 * selectively indicate certain variables can never be x using these switches:
 * - \b +xprop-nox-attr=<name>: Allow an attribute on a variable declaration 
 *   to specify the variable will never be x. 'name' is the attribute name 
 *   that will be used to indicate the variable will never be x. For 
 *   instance if the switch +xprop-nox-attr=NOX is used then a variable 
 *   tag as non-x by:
 * \code
 * (* NOX *) reg v1;          // indicate v1 is never x
 * (* NOX *) wire w1, w2, w3; // indicate w1, w2 and w3 will never be x</pre>
 * \endcode
 *
 * - \b +xprop-int-nox Declare that all variables specified as integer will 
 *   never be x. This is useful when the local design rules only use 
 *   integers as index in loops, etc.
 *
 * The \b +xprop-nox-attr=<name> switch may also be used to disable
 * x-propagation on selected constructs:
 *
 * - \b if
 * \code
 * // In the case below the attribute indicates that this if statement
 * // should not be xprop'd.
 * (* NOX *) if( condition ) ...
 * \endcode
 *
 * - \b case
 * \code
 * // In the case below the attribute indicates that this case statement
 * // should not be xprop'd.
 * (* NOX *) case( condition ) ...
 * \endcode
 *
 * - \b procedural blocks
 * \code
 * // In the case below the attribute indicates that all statements
 * // between the begin/end pair should not be xprop'd.
 * (* NOX *) begin
 * 	...
 * end
 * \endcode
 *
 * - \b structural blocks
 * \code
 * // In the case below the attribute indicates that all statements
 * // between the generate/endgenerate pair should not be xprop'd.
 * (* NOX *) generate
 * 	...
 * endgenerate
 * \endcode
 *
 * - \b functions
 * \code
 * // In the case below the attribute indicates that the function 
 * // should not be xprop'd.
 * (* NOX *) function funcName;
 * 	...
 * endfunction
 * \endcode
 *
 * - \b task
 * \code
 * // In the case below the attribute indicates that the task 
 * // should not be xprop'd.
 * (* NOX *) task funcName;
 * 	...
 * endtask
 * \endcode
 *
 * - \b module
 * \code
 * // In the case below the attribute indicates that the module 
 * // should not be xprop'd.
 * (* NOX *) task funcName;
 * 	...
 * endmodule
 * \endcode
 *
 * - \b ternary operator
 * \code
 * // In the case below the attribute indicates that the ternary operator
 * // should not be xprop'd.
 * initial out = cond ? (* NOX *) true_expression : false_expression;
 * \endcode
 *
 * - \b clock blocks 
 * \code
 * // In the case below the attribute indicates that the clock 
 * // should not be xprop'd.
 * if (* NOX *) @(posedge clock) ...
 * \endcode
 *
 *
 * These switches alter other miscellaneous behaviors:
 * - \b +xprop-allow-casex Allow casex statements in instrumented code. 
 *   Normally casex statements are disallowed due to the fact that an 
 *   'x' is treated as a true don't care in the matching criteria, 
 *   making it no possible to instrument properly.
 * - \b +xprop-begin=<pragma> Supply a comment to append before the insertion 
 *   of xprop instrumentation
 * - \b +xprop-end=<pragma> Supply a comment to post pend after the insertion 
 *   of xprop instrumentation
 * - \b +xprop-clk-begin=<pragma> Supply a comment to append before 
 *   the insertion of clock xprop instrumentation
 * - \b +xprop-clk-end=<pragma> Supply a comment to post pend after 
 *   the insertion of clock xprop instrumentation
 * - \b +xprop-clk-edge-control=[!]<tickdefine> Supply a preprocessor 
 *   tickdefine to enable instrumentation on both edges of the clock. 
 *   ! indicates tickdefine disables both edges. If switch isn't supplied 
 *   the behavior is single edge unless XPROP_BOTH_EDGES is defined.  
 *
 * Note these constructs are useful for either visually flagging xprop code 
 * or to enable other tools like coverage or synthesis to recognize the 
 * instrumentation. For example when the following switches 
 * +xprop-begin="coverage off" +xprop-end="coverage on" are used on 
 * this code snippet:
 * \code
 * if( cond ) begin
 *     a = b;
 * end else begin
 *     a = b+1;
 * end
 * \endcode
 *
 * This is the resultant output:
 * \code
 * if(cond) begin
 *    a = b;
 * end else if(~cond) begin
 *    a = b+1;
 * end
 * // coverage off
 * else begin
 *    a = 1'hx;
 * end
 * // coverage on
 * \endcode
 *
 *
 * \page xpropRational Xprop Rational
 * Xprop is a tool that instruments verilog code such that x's are 
 * propagated and not squashed. This is desirable as it reduces the 
 * simulation differences between RTL and gate level simulations. This 
 * sometimes masks bugs in RTL simulations and thereby requires a much 
 * greater degree of gate level simulation be performed. It is the 
 * experience of many who have used this tool that most of the discovered 
 * bug at the gate level would have been caught at the RTL simulation 
 * level if an xprop tool had been applied prior to RTL simulation.
 *
 * X's occur in simulation for a number of reasons:
 * - Use in RTL to indicated don't care degrees of freedom to synthesis.
 * - Uninitialized flops.
 *
 * \section Terms
 * - <b>X-propagation</b> is the process of outputing an X from a block of 
 *   logic when it's inputs are X.
 * - <b>X-pessimism</b> Is the condition when output of a block of logic 
 *   is X in cases where it never could be based upon the inputs.
 * - <b>X-optimism or X-squashing</b> Is the condition when a block of 
 *   logic outputs a non-X value when one or more of it's inputs are X. 
 *   Note this  naturally occurs in verilog due to the fact that X's are 
 *   interpreted as 0's in conditional expressions.
 *
 *
 * \section theory_of_operation Theory of Operation
 * Here is an example:
 * \code
 * if( select )
 *     // this branch is taken when cond is 1
 *     out = in1;
 * else
 *     // this branch is taken when cond is either 0 or X
 *     out = in2;
 * \endcode
 *
 * In this case an X on signal select does not result in out being X. 
 * Hence X's are squashed. Synthesis may result in code that looks 
 * like this:
 * \code
 * out = (in1 & select) | (in2 & select)
 * \endcode
 *
 * The simulation of these statements will not match in all cases.
 * Another example where gate level simulation may not match RTL is 
 * for a 2-to-1 mux expressed as:
 * \code
 * out = select ? in1 : in2;
 * \endcode
 *
 * However synthesis may result in the following:
 * \code
 * out = (in1 & select) | (in2 & select)
 * \endcode
 *
 * When select is X the simulation of these statements will differ.
 *
 * Similar issues arise with case statements and assignments to variables 
 * that use non-constant indices. The solution is to instrument the 
 * verilog code to add pessimism to such statements.
 * 
 * One way to model this is whenever an ambitious (relative to X) construct 
 * is used there are multiple possible outcomes. These multiple 
 * possible outcomes need to be modeled resolving the ambiguity create by the X.
 * This can be done using a XResolution() function. It takes two inputs, 
 * one for each possible expression output. The resolution function combines 
 * the two possiblities and creates a pessimistic output. For instances 
 * the code:
 * \code
 * if( select )
 *     out = in1;
 * else
 *     out = in2;</pre>
 * \endcode
 * Can be abstractly represented as:
 * \code
 * if( ^select === 1'bX )
 *     out = XResolution( in1, in2 );
 * else if( select )
 *     out = in1;
 * else
 *     out = in2;
 * \endcode
 *
 * Vrq uses a very simplistic XResolution function that results in less 
 * code bloat and overhead at the expense of increased pessimism. The 
 * XResolution function ignores the inputs and always returns X.
 *
 *
 * \section transformations Transformations
 *
 * \subsection if_transformation If Transformation
 * The following code snippet:
 * \code
 * if( cond ) being
 *     r1 = r2;
 * end
 * \endcode
 * Is transformed into:
 * \code
 * if(cond) begin
 *     r1 = r2;
 * end else if(~cond) ;
 * else begin
 *     r1 = 1'hx;
 * end
 * \endcode
 *
 * When an else clause is included, this code snippet:
 * \code
 * if( cond ) begin
 *     r1 = r2;
 * end else begin
 *    r1 = r3;
 * end
 * \endcode
 * is transformed into:
 * \code
 * if(cond) begin
 *     r1 = r2;
 * end else if(~cond) begin
 *     r1 = r3;
 * end else begin
 *     r1 = 1'hx;
 * end
 * \endcode
 *
 * \section case_transformation Case Constructs
 * A case statement without a default clause, like this:
 * \code
 * case( cond )
 * 0: out = in0;
 * 1: out = in1;
 * 2: out = in2;
 * endcase
 * \endcode
 *
 * Is transformed into this:
 * \code
 * case(cond)
 * 0: begin
 *     out = in0;
 *    end
 * 1: begin
 *     out = in1;
 *    end
 * 2: begin
 *     out = in2;
 *    end
 * default: begin
 *    case(^cond)
 *        1'b0,
 *        1'b1: ;
 *           default: begin
 *              out = 1'hx;
 *           end
 *     endcase
 *   end
 * endcase
 * \endcode
 * 
 * A case statement with a default clause is transformed from:
 * \code
 * case( cond )
 * 0: out = in0;
 * 1: out = in1;
 * 2: out = in2;
 * default: out = in3;
 * endcase
 * \endcode
 * to this:
 * \code
 * case(cond)
 * 0: begin
 *     out = in0;
 * end
 * 1: begin
 *     out = in1;
 * end
 * 2: begin
 *     out = in2;
 * end
 * default: begin
 *     case(^cond)
 *         1'b0,
 *         1'b1: begin
 *     out = in3;
 *     end
 *     default: begin
 *             out = 1'hx;
 *         end
 *     endcase
 * end
 * endcase
 * \endcode
 *
 *
 * A case statement with a constant case expression is transformed from:
 * \code
 * case( 1'b1 )
 * cond1: out = in0;
 * cond2: out = in1;
 * cond3: out = in2;
 * default: out = in3;
 * endcase
 * \endcode
 * to this:
 * \code
 * case(^{cond1,cond2,cond3})
 * 1'b0,
 * 1'b1: begin
 *     case(1'b0)
 *         cond1: begin
 *             out = in0;
 *         end
 *         cond2: begin
 *             out = in1;
 *         end
 *         cond3: begin
 *             out = in2;
 *         end
 *         default: begin
 *             out = in3;
 *         end
 *     endcase
 * end
 * default: begin
 *     out = 1'hx;
 * end
 * endcase
 * \endcode
 * \section bit_select_transformation Assignment with Bit Select
 * An assignment with a bit select is transformed from:
 * \code
 * reg [3:0] out;
 * ...
 * out[i] = in;
 * \endcode
 * to:
 * \code
 * case(^i)
 * 1'b0,
 * 1'b1: begin
 *      out[i] = in;
 * end
 * default: begin
 *      out = 4'hx;
 * end
 * endcase
 * \endcode
 *
 * \section part_select_transformation Assignment with Part Select
 * An assignment with a part select is transformed from:
 * \code
 * reg [1:0] in;
 * reg [3:0] out;
 * ...
 * out[i+:2] = in;
 * \endcode
 * to:
 * \code
 * case(^i)
 * 1'b0,
 * 1'b1: begin
 *     out[i+:2] = in;
 * end
 * default: begin
 *     out = 4'hx;
 * end
 * endcase
 * \endcode
 *
 * \section array_with_index_transformation Assignment to Array with Index
 * The following code gets transformed:
 * \code
 * reg [3:0] in;
 * reg [3:0] array[8:0][2:0];
 * ...
 * array[w][y] = in;
 * array[w][1] = in;
 * array[w][y][1:0] = in[1:0];
 * \endcode
 * to:
 * \code
 * // array[w][y] = in;
 * case(^{w,y})
 * 1'b0,
 * 1'b1: begin
 *     array[w][y] = in;
 * end
 * default: begin : ___block0002
 *     integer ___index0000;
 *     integer ___index0001;
 *     for(___index0001 = 2&gt;=0 ? 0 : 2;___index0001&lt;=(2&gt;=0 ? 2 : 0);___index0001 = ___index0001+1) begin
 *         for(___index0000 = 8&gt;=0 ? 0 : 8;___index0000&lt;=(8&gt;=0 ? 8 : 0);___index0000 = ___index0000+1) begin
 *              array[___index0000][___index0001] = 4'hx;
 *         end
 *     end
 * end
 * endcase
 *
 * // array[w][1] = in;
 * case(^w)
 * 1'b0,
 * 1'b1: begin
 *      array[w][1] = in;
 * end
 * default: begin : ___block0004
 *     integer ___index0003;
 *     for(___index0003 = 8&gt;=0 ? 0 : 8;___index0003&lt;=(8&gt;=0 ? 8 : 0);___index0003 = ___index0003+1) begin
 *          array[___index0003][1] = 4'hx;
 *     end
 * end
 * endcase
 *
 * // array[w][y][1:0] = in[1:0];
 * case(^{w,y})
 * 1'b0,
 * 1'b1: begin
 *      array[w][y][1:0] = in[1:0];
 * end
 * default: begin : ___block0007
 *     integer ___index0005;
 *     integer ___index0006;
 *     for(___index0006 = 2&gt;=0 ? 0 : 2;___index0006&lt;=(2&gt;=0 ? 2 : 0);___index0006 = ___index0006+1) begin
 *          for(___index0005 = 8&gt;=0 ? 0 : 8;___index0005&lt;=(8&gt;=0 ? 8 : 0);___index0005 = ___index0005+1) begin
 *              array[___index0005][___index0006][1:0] = 2'hx;
 *          end
 *      end
 *  end
 *  endcase
 * \endcode
 * 
 * \section hook_transformation ? Operator
 * This code is transformed:
 * \code`
 * out = cond ? in1 : in2;
 * \endcode
 * to:
 * \code
 * out = cond ? in1 : ~cond ? in2 : 1'hx;
 * \endcode
 * 
 * \section clock_enable_t Clock Enables
 * This code is transformed:
 * \code
 * always @(posedge clk)
 *     r <= r_P;
 * \endcode
 * to:
 * \code
 * `ifdef XPROP_BOTH_EDGES
 * always @(clk)
 * `else
 * always @(posedge clk)
 * `endif
 * if( clk ) begin
 *     r <= r_P;
 * end else if( ~clk ) ; 
 * else begin
 *     r <= 1'bx;
 * end
 * \endcode
 */

/**
 * \cond PLUGIN_INTERNALS
 */

#include <vector>
#include <fstream>
#include <set>
#include "glue.h"
#include "main.h"
#include "cxprop.h"
#include "cdumpverilog.h"
#include "cmodule.h"
#include "cport.h"
#include "cfref.h"
#include "cvar.h"
#include "cnet.h"
#include "cparam.h"
#include "cnode.h"
#include "cinstance.h"
#include "cfunction.h"
#include "cportdir.h"
#include "array.h"

enum    TranslateState { eTRANSLATE_DEFAULT, eTRANSLATE_OFF, eTRANSLATE_ON } ;
static  TranslateState	translateState;
static  CNode* prependToModule;
static  CNode* insertionBeginComment = NULL;
static  CNode* insertionEndComment = NULL;
static  CNode* insertionClkBeginComment = NULL;
static  CNode* insertionClkEndComment = NULL;
static  int disableIf = 0;
static  int disableCase = 0;
static  int disableSelect = 0;
static  int disableArray = 0;
static  int disableTernary = 0;
static  char* noxAttribute = NULL;
static  int integerIsNoX = 0;
static  int enableClock  = 0;
static  const char* edgeControl  = NULL;
static  const char* edgeAttrName  = NULL;
static  Message* mCCBMS;
static  Message* mCDWOO;
static  Message* mNSTOS;
static  Message* mUESCI;

struct LHS {
	CNode*   ref;
	CNode*   delay;
	NodeOp_t op;
	Coord_t* loc;
	vector<CNode*> dim;
	CNode*	 msb;
	CNode*   lsb;
	int	 psType;
	int      hasBitSelect;
	int	 xOutPartSelect;
	vector<int> xOutDim;
	// only used for later stage processing
        set<int> coveredBitVector;
	int	 coveringInfoValid;
	int	 fullyCovered;
	int	 drop;
	// methods
	unsigned Hash() {
	    unsigned result = 0; 
	    int s1 = dim.size() - (hasBitSelect ? 1 : 0);
	    if( ref )   { result ^= ref->Hash(); }
	    if( delay ) { result ^= delay->Hash(); }
	    result ^= s1;
	    for( int i = 0; i < s1; ++i ) {
		if( dim[i] ) { result ^= dim[i]->Hash(); }
	    }
	    return result;
	}
	void Dump() {
	    printf( "LHS: " );
	    printf( "ref=%p,delay=%p,op=%d,msb=%p,lsb=%p"
		    ",psType=%d,hasBitSelect=%d,dsize=%d,ciValid=%d,fully=%d"
		    ",covered={", 
		    ref, delay, op, msb, lsb, psType, hasBitSelect, dim.size(),
		    coveringInfoValid, fullyCovered  );
	    set<int>::iterator ptr;
	    for( ptr = coveredBitVector.begin(); 
		 ptr != coveredBitVector.end(); ++ptr ) {
		if( ptr != coveredBitVector.begin() ) {
		    printf( "," );
		}
		printf( "%d", *ptr );
	    }
	    printf( "}\n" );
	}
	int Compatible( LHS* lhs )
	{
	    int s1 = dim.size() - (hasBitSelect ? 1 : 0);
	    int s2 = lhs->dim.size() - (lhs->hasBitSelect ? 1 : 0);
            if( !Equivalent( ref, lhs->ref ) || 
                !Equivalent( delay, lhs->delay ) || 
                op != lhs->op || 
                s1 != s2 ) { 
	        return FALSE;
            }
            int indexMismatch = 0;
            for( int d = 0; d < s1; ++d ) {
                if( !Equivalent( dim[d], lhs->dim[d] )  ) {
	    	return FALSE;
                }
            }

	    /*
 	     * if covering info is available and vector is fully
 	     * covered, then entries are compatible.
 	     */
	    if( (coveringInfoValid && fullyCovered) ||
	        (lhs->coveringInfoValid && lhs->fullyCovered) ) {
		return TRUE;
	    }
	    /*
 	     * if covering info is avalable for both entrys
 	     * they are compatable and can match. Otherwise
 	     * the selects must match exactly.
 	     */
	    if( coveringInfoValid != lhs->coveringInfoValid ) {
	        return FALSE;
	    }
	    if( !coveringInfoValid && !lhs->coveringInfoValid ) {
	        if( hasBitSelect != lhs->hasBitSelect ) {
	    	return FALSE;
	        }
    	        if( hasBitSelect && lhs->hasBitSelect &&
    	        	!Equivalent( dim.back(), lhs->dim.back()) ) {
    	            return FALSE;
    	        }
	        if( psType != lhs->psType ) {
	    	return FALSE;
	        }
    	        if( msb && !Equivalent( msb, lhs->msb) ) {
    	            return FALSE;
    	        }
    	        if( lsb && !Equivalent( lsb, lhs->lsb) ) {
    	            return FALSE;
    	        }
	    }
	    
	    return TRUE;
        }
	int IsSubsetOf( LHS* parent )
	{
	    /*
 	     * LHS must be compatible to be subset
 	     */
	    if( !Compatible( parent ) ) {
		return FALSE;
	    }
    	    /*
     	     * if covering info is not available then 
     	     * match is exact subset
     	     */
    	    if( !coveringInfoValid ) {
                    return TRUE;	   
    	    }
	
	    /*
 	     * if vector is fully covered, exact match
 	     */
	    if( parent->fullyCovered ) {
	        return TRUE;
	    }

	    /*
 	     * check if range is a subset of existing range
 	     */
	    if( !fullyCovered ) {
	        set<int> difference;
	        set_difference( coveredBitVector.begin(),
	        		coveredBitVector.end(),
	        		parent->coveredBitVector.begin(),
	        		parent->coveredBitVector.end(),
	        		inserter(difference, difference.begin() ) );
	        if( difference.empty() ) {
	            return TRUE;
	        }
	    }
	    return FALSE;
	}
};

extern  char*		outputPath;

static CNode* XPropStatement( CBlock* module, CBlock* block, CNode* n );
static CNode* CreateLHSAssignments( CBlock* module, CBlock* block,
			vector<LHS>& lhsList );
static CNode* CreateAssignment( CBlock* module, NodeOp_t op, 
	    CNode* delay, CNode* target, vector<CNode*>& dim,
	    int hasBitSelect, int psType, CNode* msb, CNode* lsb );

extern CDumpVerilog*	dumpVerilog;
extern int		errorCount;

inline CNode* Clone( CNode* n ) {
	if( n == NULL ) {
	    return NULL;
	}
	return n->Clone();
}

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
extern "C" {
CBackend* CreateToolInstance()
{
    return new CXProp();
}
}

/************************************************
 	Constructor
 ************************************************/
CXProp::CXProp( ) 
{
	translateState = eTRANSLATE_DEFAULT;
	switches.push_back( "+xprop-begin=<pragma>" );
	switchDescription["+xprop-begin=<pragma>"] = 
		"insertion begin pragma";
	switches.push_back( "+xprop-end=<pragma>" );
	switchDescription[ "+xprop-end=<pragma>" ] =
		"insertion end pragma";
	switches.push_back( "+xprop-clk-begin=<pragma>" );
	switchDescription["+xprop-clk-begin=<pragma>"] = 
		"insertion clock begin pragma";
	switches.push_back( "+xprop-clk-end=<pragma>" );
	switchDescription[ "+xprop-clk-end=<pragma>" ] =
		"insertion clock end pragma";
	switches.push_back( "+xprop-disable-if" );
	switchDescription[ "+xprop-disable-if" ] = "disable if instrumentation";
	switches.push_back( "+xprop-disable-case" );
	switchDescription[ "+xprop-disable-case" ] = 
						"disable case instrumentation";
	switches.push_back( "+xprop-disable-select" );
	switchDescription[ "+xprop-disable-select" ] = 
				"disable bit/part select instrumentation";
	switches.push_back( "+xprop-disable-array" );
	switchDescription[ "+xprop-disable-array" ] = 
					"disable array instrumentation";
	switches.push_back( "+xprop-disable-ternary" );
	switchDescription[ "+xprop-disable-ternary" ] = 
					"disable '?' instrumentation";
	switches.push_back( "+xprop-nox-attr=<name>" );
	switchDescription[ "+xprop-nox-attr=<name>" ] = 
					"attribute specifing variable will never be x";
	switches.push_back( "+xprop-int-nox" );
	switchDescription[ "+xprop-int-nox" ] = 
					"declare integers will never be x";

	switches.push_back( "+xprop-enable-clock" );
	switchDescription[ "+xprop-enable-clock" ] = 
					"instrument register clocks";

 	switches.push_back( "+xprop-clk-edge-control=[!]<tickdefine>" );
 	switchDescription[ "+xprop-clk-edge-control=[!]<tickdefine>" ] = 
 		"Supply a preprocessor tick define to enable instrumentation"
		" on both edges of the clock. ! indicates tickdefine"
		" disables both edges. If switch isn't supplied the behavior is"
		" single edge unless XPROP_BOTH_EDGES is defined";

	/*
 	 * register messages
 	 */
	mCCBMS = Message::RegisterWarning( this, Message::eERROR,
		"CCBMS", "casex cannot be made safe, please use casez" );
	mCDWOO = Message::RegisterWarning( this, Message::eERROR,
		"CDWOO", "cannot determine width of ?"
			 " operands, cannot instrument expression" );
	mNSTOS = Message::RegisterError( this, "NSTOS", 
			"nested synopsys translate_off statements" );
	mUESCI = Message::RegisterWarning( this, Message::eWARNING, "UESCI", 
			"unsupported edge sensitive construct for clock"
			" instrumentation" );
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CXProp::GetToolName( void )
{
	static char	toolName[] = "xprop";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CXProp::GetToolDescription( void )
{
	static char	toolDescription[] = "Add X propagation instrumentation";

	return toolDescription;
}

/*******************************************************
	IsTranslateOn
	- return TRUE if PRAGMA node is synopsys translate on 
********************************************************/

static int IsTranslateOn( CNode* n )
{
	char* comment = n->Arg<char*>(0);
	char* ptr = strstr( comment, "synopsys" );
	if( ptr == NULL ) {
		return FALSE;
	}
	ptr = strstr( ptr, "translate_on" );
	return (ptr != NULL );
}

/*******************************************************
	IsTranslateOff
	- return TRUE if PRAGMA node is synopsys translate off 
********************************************************/

static int IsTranslateOff( CNode* n )
{
	char* comment = n->Arg<char*>(0);
	char* ptr = strstr( comment, "synopsys" );
	if( ptr == NULL ) {
		return FALSE;
	}
	ptr = strstr( ptr, "translate_off" );
	return (ptr != NULL );
}
/*******************************************************
 	IsNoxNode
	- return TRUE if xprop is disabled for node 
	  and it's subtrees.
 ******************************************************/
static int IsNoxNode( CNode* n ) 	
{
    switch( n->GetOp() ) {
    case eBLOCK_REF:
	return  n->HasAttribute( noxAttribute );
    case eMODULE_DEF:
    case eFUNCTION_DEF: 
	return n->Arg<CBlock*>(0)->HasAttribute( noxAttribute );
    } 
    return FALSE;
}

/*******************************************************
	StringList2Comments
	- Convert list of strings to a tree of COMMENT
********************************************************/
static CNode* StringList2Comments( list<string>& args )
{
    CNode* result = NULL;
    list<string>::iterator ptr;
    for( ptr = args.begin(); ptr != args.end(); ++ptr ) {
	result = cLINK( result, cCOMMENT( strdup(ptr->c_str()) ) );	
    }
    return result;
}

/*******************************************************
	PSWidthExpression
	- return part select width expression
********************************************************/

static CNode* PSWidthExpression( CDecl* decl, CNode* msb, CNode* lsb )
{
    CNode* width;
    if( !decl->GetMsb() ) {
        width = cINT32(1);
    } else if( decl->GetMsb()->IsVolatile() ||
	       decl->GetLsb()->IsVolatile() || decl->WidthDirection() == 0 ) {
	width = cHOOK( cGE(msb->Clone(),lsb->Clone()),
                       cADD(cSUB(msb->Clone(), lsb->Clone()),cINT32(1)),
                       cADD(cSUB(lsb->Clone(), msb->Clone()),cINT32(1))
		);
			
    } else if( decl->WidthDirection() >= 0 ) {
        width = cADD(cSUB(msb->Clone(),
        		lsb->Clone()),cINT32(1));
    } else {
        width = cADD(cSUB(lsb->Clone(),
        		msb->Clone()),cINT32(1));
    }

    return width;
}

/*******************************************************
	XPropCaseItem
	- add x propagation logic 
********************************************************/

static CNode* XPropCaseItem( CBlock* module, CBlock* block, CNode* n ) 
{
        if( n->GetOp() == eCOMMENT ||
	    n->GetOp() == eVRQ ||
	    n->GetOp() == ePRAGMA ) {
                return n;
        }
	CNode* orig = n;
        while( n->GetOp() == eLIST ) {
            n->Arg<CNode*>(0) = XPropCaseItem( module, block, 
						n->Arg<CNode*>(0) ); 
		
	    if( n->Arg<CNode*>(1)->GetOp() != eLIST ) {
                n->Arg<CNode*>(1) = 
			XPropCaseItem( module, block, n->Arg<CNode*>(1) );
                return orig;
	    }
	    n = n->Arg<CNode*>(1);
        }

        n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
	return n;

}
/*******************************************************
	_CollectIndex
	- collect index expression from bit selects
	  in vector 
********************************************************/

struct Pair {
	// reference to decl
	CNode* 	     target;
	// 1 if last index is a bit/part select
	int hasBitSelect;
	// -1 for -:, 0 for :, 1 for +:
	int psType;
	// part select expressions
	CNode* msb;
	CNode* lsb;
	// list of indexes including bit selects or variable part selects
	vector<CNode*> dim;
};

static void CollectDimIndexes( CNode* n, vector<CNode*>& list )
{
	if( !n ) {
	    return;
	}
	switch( n->GetOp() ) {
	case eCOMMENT:
	case eVRQ:
	case ePRAGMA:
	    return;
	case eELIST:
	    while(1) {
	        CollectDimIndexes( n->Arg<CNode*>(0), list );
		n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eELIST ) {
	    	    CollectDimIndexes( n, list );
		    break;
		}
	    }
	    return;
	default: 
	    list.push_back( n );
	    break;
	}
}

static int _CollectIndex( CNode* n, void* arg )
{
	vector<Pair>* list = (vector<Pair>*)arg;
	switch( n->GetOp() ) {
	case eVAR_REF:
	    {
		if( n->Arg<CVar*>(0)->GetNodeType() != eS &&
		    n->Arg<CVar*>(0)->GetNodeType() != eB ) {
		    return TRUE;
		}
		Pair p;
	        p.target = n;
		p.hasBitSelect = 0;
		p.psType = 0;
		p.msb = NULL;
		p.lsb = NULL;
		list->push_back( p );
	    }
	    break;
	case eARRAY:
	    if( n->Arg<CNode*>(0)->GetOp() == eEXTERNAL_REF ) { 
		return TRUE;
	    }
	    {
		Pair p;
		p.target = n->Arg<CNode*>(0);
	        CollectDimIndexes( n->Arg<CNode*>(1), p.dim );
		MASSERT( p.target->GetOp() == eVAR_REF );
		int hasSelect = p.dim.size() > p.target->Arg<CVar*>(0)->GetNumberOfUnpackedDimensions();
	        CNode* select = NULL;
		if( hasSelect ) {
	            select = p.dim.back();
		    p.dim.pop_back();
		}
	        if( disableArray ) {
		    p.dim.erase( p.dim.begin(), p.dim.end() );
	        }
		p.hasBitSelect = 0;
		p.psType = 0;
		p.msb = NULL;
		p.lsb = NULL;
		CNode* selectSave = select;
		while( select && select->GetOp() == eMACRO_EXPR ) {
		   select = select->Arg<CNode*>(1);
		}
		// ??? mdh - should treat this as a warning if macro is null
		if( !select ) {
		   select = selectSave;
		}
		if( hasSelect && !disableSelect ) {
		    switch( select->GetOp() ) {
		    case eSLICE:
			p.psType = 0;
			p.msb = select->Arg<CNode*>(0);
			p.lsb = select->Arg<CNode*>(1);
			break;
		    case ePSLICE:
			p.psType = 1;
			p.msb = select->Arg<CNode*>(0);
			p.lsb = select->Arg<CNode*>(1);
			break;
		    case eMSLICE:
			p.psType = -1;
			p.msb = select->Arg<CNode*>(0);
			p.lsb = select->Arg<CNode*>(1);
			break;
		    default:
		        p.hasBitSelect = 1;
		        p.dim.push_back( select );
			break;
		    }
		}
		list->push_back( p );
	    }
	    return FALSE;
	default: ;
	}
	return TRUE;
}
/*******************************************************
	IsNonX
	- return TRUE if expression cannot resolve to X
********************************************************/
static int IsNonX( CNode* n )
{
    switch( n->GetOp() ) {
    case eLIST:
	return IsNonX( n->Arg<CNode*>(0) ) && IsNonX( n->Arg<CNode*>(1) );
    default:
	break;
    }
    return n->IsNonX(integerIsNoX,noxAttribute) || n->GetNodeType() == eR;
}
/*******************************************************
	_CollectTargets
	- collect lhs of statements 
********************************************************/

struct TargetArg {
	vector<LHS>* list;
	CNode*       delay;
	NodeOp_t     op;
	Coord_t*     loc;
	CNode*	     dim;
	CNode*	     msb;
	CNode*	     lsb;
	int	     psType;
	int	     hasBitSelect;
};
static TranslateState lhsTranslateState;

static int _CollectRHD( CNode* n, void* arg )
{
	list<CDecl*>* l = (list<CDecl*>*) arg;

	switch( n->GetOp() ) {
	case eVAR_REF:
	case ePARAM_REF:
	case eNET_REF:
	case ePORT_REF:
	case eEXTERNAL_REF:
	case eGENVAR_REF:
		l->push_back(n->Arg<CDecl*>(0) );
		break;
	default:
		return TRUE;
	}
	return FALSE;
}

static int _CollectTargets( CNode* n, void* arg )
{
	TargetArg* ta = (TargetArg*) arg;

	switch( n->GetOp() ) {
	case eVAR_REF:
		if( n->Arg<CVar*>(0)->GetNodeType() != eS &&
		    n->Arg<CVar*>(0)->GetNodeType() != eB ) {
		    return TRUE;
		}
		{
			LHS lhs;
			lhs.ref = n;
			lhs.delay = ta->delay; 
			lhs.op = ta->op;
			lhs.loc = ta->loc;
			lhs.msb = ta->msb; 
			lhs.lsb = ta->lsb;
			lhs.psType = ta->psType;  
			lhs.hasBitSelect = ta->hasBitSelect;
			lhs.xOutPartSelect = 0;
			if( ta->dim ) {
	                    CollectDimIndexes( ta->dim, lhs.dim );
			}
			if( ta->hasBitSelect ) {
			    lhs.dim.push_back(ta->msb);
			    lhs.msb = NULL;
		            lhs.lsb = NULL;
			}
			if( ta->dim ) {
			    lhs.hasBitSelect = lhs.dim.size() > 
				n->Arg<CVar*>(0)->GetNumberOfUnpackedDimensions();
			}
			lhs.xOutDim.assign(lhs.dim.size(),0);
			ta->list->push_back( lhs );
		}
		break;
	case eARRAY: {
		if( n->Arg<CNode*>(0)->GetOp() == eEXTERNAL_REF ) {
		    return TRUE;
		}
		ta->dim = n->Arg<CNode*>(1);
		vector<CNode*> indexes;
		EList2VectorExclude( ta->dim, set<NodeOp_t>(), indexes );
		CNode* last = indexes.back();
		while( last && last->GetOp() == eMACRO_EXPR ) {
		    last = last->Arg<CNode*>(1);
		}
		// ??? mdh - this code currently will treat an index
		// with an undefined macro as a bit select. Should
		// issue a warning.
		if( !last ) {
		    last = indexes.back();
		}
		switch( last->GetOp() ) {
		case eSLICE:
		    ta->psType = 0;
		    ta->msb = last->Arg<CNode*>(0);
		    ta->lsb = last->Arg<CNode*>(1);
		    indexes.pop_back();
		    ta->dim = Vector2EList( indexes );
		    break;
		case ePSLICE:
		    ta->psType = 1;
		    ta->msb = last->Arg<CNode*>(0);
		    ta->lsb = last->Arg<CNode*>(1);
		    indexes.pop_back();
		    ta->dim = Vector2EList( indexes );
		    break;
		case eMSLICE:
		    ta->psType = -1;
		    ta->msb = last->Arg<CNode*>(0);
		    ta->lsb = last->Arg<CNode*>(1);
		    indexes.pop_back();
		    ta->dim = Vector2EList( indexes );
		    break;
		default: 
		    ta->hasBitSelect = (indexes.size() > Array::ArrayDecl( n->Arg<CNode*>(0) )->GetNumberOfUnpackedDimensions());
		    if( ta->hasBitSelect ) {
		        ta->msb = indexes.back();
		        ta->lsb = indexes.back();
			indexes.pop_back();	
		        ta->dim = Vector2EList( indexes );
		    }
		    break;
		}
		n->Arg<CNode*>(0)->PreVisit1( _CollectTargets, arg );
		ta->psType = 0;
		ta->hasBitSelect = 0;
		ta->msb = NULL;
		ta->lsb = NULL;
		ta->dim = NULL;
		} break;
	default:
		return TRUE;
	}
	return FALSE;
}
/*******************************************************
	_CollectLHS
	- collect lhs of statements 
********************************************************/

static int _CollectLHS( CNode* n, void* arg )
{
	vector<LHS>* lhsList = (vector<LHS>*) arg;
	switch( n->GetOp() ) {
	case ePRAGMA:
		if( IsTranslateOn(n) ) {
			lhsTranslateState = eTRANSLATE_ON;
		} else if( IsTranslateOff(n) ) {
			if( lhsTranslateState == eTRANSLATE_OFF ) {
				message( n->GetCoord(), mNSTOS );
			}
			lhsTranslateState = eTRANSLATE_OFF;
		}
		break;
	case eASSIGN:
	case eNBASSIGN:
		{
		    if( lhsTranslateState != eTRANSLATE_OFF ) {
			CNode* l = n->Arg<CNode*>(0);
			CNode* rl = n->Arg<CNode*>(1);
			TargetArg ta = { (vector<LHS>*)arg, l, n->GetOp(), 
				n->GetCoord(), NULL, NULL, NULL, 0, 0 };
			rl->PreVisit1( _CollectTargets, &ta );		
		    }
		}
		return FALSE;
	default:
		return TRUE;
	}
}
/*******************************************************
	_CountEdges
	- count the number of edge constructs in expression
********************************************************/

static void _CountEdges( CNode* n, void* arg )
{
    int* count = (int*)arg;

    switch( n->GetOp() ) {
    case ePOSEDGE:
    case eNEGEDGE:
    case eEDGE:
	(*count)++;
	break;
    }
}
/*******************************************************
	_CollectCaseDefault
	- find ptr to default CASEITEM
	  if found set *arg to caseitem node pointer
********************************************************/

static int _CollectCaseDefault( CNode* n, void* arg )
{
	if( n->GetOp() == eCASEITEM ) {
		if( n->Arg<CNode*>(0) == NULL ) {
		    *(CNode**)arg = n;
		}
		return FALSE;
	}
	return TRUE;
}

/*******************************************************
	_CollectCases
	- collect expression from case item in case statements
********************************************************/

static int _CollectCases( CNode* n, void* arg )
{
	list<CNode*>& caseList = *(list<CNode*>*)arg;
	switch( n->GetOp() ) {
	case eCASEITEM:
	    if( n->Arg<CNode*>(0) && 
	        !IsNonX(n->Arg<CNode*>(0)) && 
	        !n->Arg<CNode*>(0)->IsConstant() ) {
	        caseList.push_back( n->Arg<CNode*>(0) );
	    }
	    return FALSE;
	default:
	    return TRUE;
	}
}
/*******************************************************
  	InScope
	- return 1 if decl is in symtab
 *******************************************************/
static int InScope( CSymtab<CDecl>& symtab, CDecl* d ) 
{
    CDecl* ld = symtab.Lookup( d->GetSymbol() );
    return ld && ld == d;

}

/*******************************************************
  	ExpressionInScope
	- return 1 if all decls in expression are in scope
 *******************************************************/
static int ExpressionInScope( CSymtab<CDecl>& symtab, CNode* n ) 
{
    list<CDecl*> RHD;
    n->PreVisit1( _CollectRHD, &RHD );		
    list<CDecl*>::iterator dptr;
    int inScope = 1;
    for( dptr = RHD.begin(); dptr != RHD.end(); ++dptr ) {
	if( !InScope( symtab, *dptr ) ) {
	    return 0;	
	}
    }
    return 1;
}
/*******************************************************
	EList2Cat
	- create cat of from list of expression
********************************************************/
static CNode* EList2Cat( CNode* n )
{
	if( !n ) {
	    return NULL;
	}
	if( n->GetOp() != eELIST ) {
	    return n->Clone();
	}
	CNode* result = NULL;
	CNode* last   = NULL;
	while(1) {
	    /*
	     * Note: n is put in to the right fork to keep
	     *       _Cat func from creating a eUCAT. It
	     *       will be overridden
	     */
	    CNode* next = cCAT( EList2Cat(n->Arg<CNode*>(0)), n );
	    if( !result ) {
		result = next;
	    }
	    if( last ) {
		last->Arg<CNode*>(1) = next;
	    }
	    n = n->Arg<CNode*>(1);
	    if( n && n->GetOp() != eELIST ) {
	        next->Arg<CNode*>(1) = EList2Cat( n );
	        break;
	    }
	    last = next;
	}
	return result;
}

/*******************************************************
	CreateXAssignment
	- create a block of code to assign x to target
********************************************************/
static CNode* CreateXAssignment( 
	    CBlock* module, 
	    CBlock* block, 
	    NodeOp_t op, 
	    CNode* delay, 
	    CNode* target,
	    vector<CNode*>& dim,
	    int hasBitSelect,
	    int psType,
	    CNode* msb,
	    CNode* lsb,
	    int conditional,
	    int xOutPartSelect,
	    vector<int>& xOutDim,
	    int forceIndicesNoX )	
{
    CNode* t = NULL;
    CDecl* decl = target->Arg<CDecl*>(0);
    int isWidthVolatile = target->IsVolatile();
    target = target->Clone();
    /*
     * only instrument non-external regs 
     */
    if( target->GetOp() == eEXTERNAL_REF )  {
	return t;
    } 
    if( target->GetOp() != eVAR_REF ) {
        return t;
    }
    if( target->Arg<CVar*>(0)->GetNodeType() != eB && 
        target->Arg<CVar*>(0)->GetNodeType() != eS ) {
        return t;
    }
    /*
     * if integers are declared to non-x, don't force them to x.
     */
    if( integerIsNoX && 
        target->Arg<CVar*>(0)->GetDataType()->GetVarDataType() == eINTEGER ) {
	return t;
    }
    /*
     * check indexes to see if array can be treated 
     * as simple vector
     */
    int indexCount = dim.size();
    int dimConstant = 1;
    if( hasBitSelect ) {
	indexCount--;
    }
    for( int i = 0; i < indexCount; ++i ) {
	if( !dim[i]->IsConstant() ) {
		dimConstant = 0;
		break;
	}	
    }
    int isArray = decl->GetNumberOfUnpackedDimensions() > 0;
    CVar* reg = target->Arg<CVar*>(0);
    CNode* s;	
    CNode* select = NULL;
    CNode* width = NULL;
    CNode* altSelect = NULL;
    CNode* altWidth = NULL;

    /*
     * assignments for targets in forloops should
     * never generate conditional assignments all indices are
     * x'd out.
     */
    int allXd = 1;
    int anyXd = 0;
    if( (hasBitSelect && !xOutDim.back()) || 
	(psType != 0 && !xOutPartSelect) ) {
	allXd = 0;
    }
    if( (hasBitSelect && xOutDim.back()) || 
	(psType != 0 && xOutPartSelect) ) {
	anyXd = 1;
    }
    vector<int>::iterator iptr;
    for( iptr = xOutDim.begin(); iptr != xOutDim.end(); ++iptr ) {
	if( *iptr == 0 ) {
	     allXd = 0;
	} else {
	     anyXd = 1;
	}
    }
// MDH - this is currently very conservative. However being less
// conservative will require some major re-write below.
    conditional &= !anyXd;

    /*
     * Handle bit selects
     */
    if( hasBitSelect && 
	(dim.back()->IsConstant() ||
	(!xOutDim.back() && IsNonX(dim.back()) ) ) ) { 
        CVector x1(1);
        x1.LoadHex( "x" );
        x1.SetPreferredBase( 16 );
        width = cVECTOR(x1);
        select = cARRAY( NULL, dim.back()->Clone() );
    /*
     * Handle non-volatile part selects
     */
    } else if( msb && lsb && psType == 0 &&
    	   msb->IsConstant() && lsb->IsConstant() ) {
        if( msb->IsVolatile() || lsb->IsVolatile() ||
	    !msb->IsEvaluateable() || !lsb->IsEvaluateable() ) {
	    width = PSWidthExpression( reg, msb, lsb );
            CVector x1(1);
            x1.LoadHex( "x" );
            x1.SetPreferredBase( 16 );
    	    width = cREP( width,cVECTOR(x1) );
        } else {
            int m = msb->EvalINT32();
            int l = lsb->EvalINT32();
            int w = (m>l) ? (m-l)+1 : (l-m)+1;
            CVector xw(w);
            xw.LoadHex( "x" );
            xw.SetPreferredBase( 16 );
            width = cVECTOR(xw);
        }
        select = cARRAY( NULL, cSLICE( msb->Clone(), lsb->Clone() ) );
    /*
     * Handle constant range selectors
     */
    } else if( psType != 0 && msb && 
	       (msb->IsConstant() ||
	       (!xOutPartSelect && IsNonX(msb) ) ) ) {
        if( lsb->IsVolatile() || !lsb->IsEvaluateable() ) {
            CVector x1(1);
            x1.LoadHex( "x" );
            x1.SetPreferredBase( 16 );
    	    width = cREP(lsb->Clone(),cVECTOR(x1));
        } else {
            int w = lsb->EvalINT32();
            CVector xw(w);
            xw.LoadHex( "x" );
            xw.SetPreferredBase( 16 );
            width = cVECTOR(xw);
        }
        if( psType > 0 ){
            select = cARRAY( NULL, cPSLICE( msb->Clone(), lsb->Clone() ) );
        } else {
            select = cARRAY( NULL, cMSLICE( msb->Clone(), lsb->Clone() ) );
        }
    /*
     * Handle default case
     */
    } else {
	width = PSWidthExpression( reg, reg->GetMsb(), reg->GetLsb() );
        if( !width->IsVolatile() && width->IsEvaluateable() ) {
           CVector xt(width->EvalINT32());
           xt.LoadHex( "x" );
           xt.SetPreferredBase( 16 );
           width = cVECTOR(xt);
        } else {
            CVector x1(1);
            x1.LoadHex( "x" );
            x1.SetPreferredBase( 16 );
    	    width = cREP(width,cVECTOR(x1));
        }
    }
    /*
     * calculate alternative width and selector if
     * conditional expression is required
     */
    if( conditional ) {
        /*
         * Handle bit selects
         */
        if( hasBitSelect ) {
            CVector x1(1);
            x1.LoadHex( "x" );
            x1.SetPreferredBase( 16 );
            altWidth = cVECTOR(x1);
            altSelect = cARRAY( NULL, dim.back()->Clone() );
        /*
         * Handle part selects
         */
        } else if( msb && psType == 0 ) { 
            if( msb->IsVolatile() || lsb->IsVolatile() ||
		!msb->IsEvaluateable() || !lsb->IsEvaluateable() ) {
	        altWidth = PSWidthExpression( reg, msb, lsb );
                CVector x1(1);
                x1.LoadHex( "x" );
                x1.SetPreferredBase( 16 );
    	        altWidth = cREP( altWidth,cVECTOR(x1) );
            } else {
                int m = msb->EvalINT32();
                int l = lsb->EvalINT32();
                int w = (m>l) ? (m-l)+1 : (l-m)+1;
                CVector xw(w);
                xw.LoadHex( "x" );
                xw.SetPreferredBase( 16 );
                altWidth = cVECTOR(xw);
            }
            altSelect = cARRAY( NULL, cSLICE( msb->Clone(), lsb->Clone() ) );
        /*
         * Handle constant range selectors
         */
        } else if( psType != 0 && msb ) { 
            if( lsb->IsVolatile() ) {
                CVector x1(1);
                x1.LoadHex( "x" );
                x1.SetPreferredBase( 16 );
    	        altWidth = cREP(lsb->Clone(),cVECTOR(x1));
            } else {
                int w = lsb->EvalINT32();
                CVector xw(w);
                xw.LoadHex( "x" );
                xw.SetPreferredBase( 16 );
                altWidth = cVECTOR(xw);
            }
            if( psType > 0 ){
                altSelect = cARRAY( NULL, cPSLICE( msb->Clone(), lsb->Clone() ) );
            } else {
                altSelect = cARRAY( NULL, cMSLICE( msb->Clone(), lsb->Clone() ) );
            }
        /*
         * Handle default case
         */
        } else {
	    altWidth = PSWidthExpression( reg, reg->GetMsb(), reg->GetLsb() );
            if( !altWidth->IsVolatile() && altWidth->IsEvaluateable() ) {
               CVector xt(altWidth->EvalINT32());
               xt.LoadHex( "x" );
               xt.SetPreferredBase( 16 );
               altWidth = cVECTOR(xt);
            } else {
                CVector x1(1);
                x1.LoadHex( "x" );
                x1.SetPreferredBase( 16 );
    	        altWidth = cREP(altWidth,cVECTOR(x1));
            }
        }

    }
    /* 
     * handle vectors 
     */
    if( !isArray ) {
	CNode* target1 = target->Clone();
	if( indexCount ) {
	    // ??? mdh - I don't this code is ever used
	    MASSERT( FALSE );
	    CNode* index = NULL;
	    for( int i = 0; i < indexCount; ++i ) {
		index = cELINK( index, dim[i]->Clone() );
	    }
	    target1 = cARRAY( target1, index );
	}
	if( select ) {
	    select->Arg<CNode*>(0) = target1;
	    target1 = select;
	}
	if( op == eASSIGN ) {
		s = cASSIGN( Clone(delay), target1, width );
	} else {
		s = cNBASSIGN( Clone(delay), target1, width );
	}
	if( conditional ) {
	    CNode* cond = NULL;
	    if( altSelect && hasBitSelect &&
	        !dim.back()->IsConstant() && 
	        (xOutDim.back() || !IsNonX(dim.back()) ) ) { 
		cond = cond ?  cCAT( cond, dim.back()->Clone() ) : 
				           dim.back()->Clone();
	    }
	    if( altSelect && psType != 0 &&
	        !msb->IsConstant() && 
	        (xOutPartSelect || !IsNonX(msb)) ) { 
		cond = cond ?  cCAT( cond, msb->Clone() ) : 
				           msb->Clone();
	    }
	    if( cond ) {
	        CVector zero(1);
	        zero.LoadBinary( "0" );
	        CVector one(1);
	        one.LoadBinary( "1" );
		CNode* alt = NULL;
		if( cond->IsWidthVolatile() || 
		    !cond->IsWidthEvaluateable() || cond->GetWidth() != 1 ) {
		    cond = cRXOR(cond);
	        }
		CNode* indexes = NULL;
                for( int i = 0; i < indexCount; ++i ) {
                    indexes = cELINK( indexes, dim[i]->Clone() );
                }
		CNode* lhs1 = target->Clone();
		if( altSelect ) {
		    MASSERT( altSelect->GetOp() == eARRAY );
		    altSelect->Arg<CNode*>(0) = lhs1;
		    altSelect->Arg<CNode*>(1) = cELINK(indexes,altSelect->Arg<CNode*>(1));
	            lhs1 = altSelect;
		} else {
		    if( indexes ) {
	                lhs1 = cARRAY( lhs1, indexes );
		    }
		}
	        if( op == eASSIGN ) {
	            alt = cASSIGN( Clone(delay), lhs1, altWidth );
	        } else {
	            alt = cNBASSIGN( Clone(delay), lhs1, altWidth );
	        }
	        s = cCASE( cond, cLIST(
		    cCASEITEM( cELIST(cVECTOR(zero),cVECTOR(one)), alt ),
		    cCASEITEM( NULL, s ) ) );
	    }
	}
	t = cLINK( s, t );
        return t;
    /*
     * Handle arrays with volatile indexes
     */
    } else {
	CVar* decl = target->Arg<CVar*>(0);
	int dimensions = decl->GetNumberOfUnpackedDimensions();
	MASSERT( dimensions >= indexCount );
	/*
	 * force whole array to x
	 */
	vector<CNode*> lb(dimensions);
	vector<CNode*> mb(dimensions);
	vector<CVar*> index(dimensions);
	vector<int> nonx(dimensions);
	Coord_t coord = { 0, "???" };
	CNode* indexes = NULL;
	CNode* declStmts = NULL;
	/*
 	 * create expression/variables for index/msb/lsb of each dimension
 	 */
        for( int i = 0; i < dimensions; ++i ) {
	    if( i >= indexCount  ) {
		nonx[i] = 0;
	    } else if( forceIndicesNoX || dim[i]->IsConstant() ||
	            (!xOutDim[i] && IsNonX(dim[i])) ) {
	            nonx[i] = 1;
	            continue;
	    } else {
	        nonx[i] = 0;
	    }
    	    lb[i] = cHOOK( cGE(decl->GetUnpackedMsi(i),
			   decl->GetUnpackedLsi(i)), 
    	    		   decl->GetUnpackedLsi(i), 
    	    		   decl->GetUnpackedMsi(i), &coord );
    	    mb[i] = cHOOK( cGE(decl->GetUnpackedMsi(i),
			   decl->GetUnpackedLsi(i)), 
    	    		   decl->GetUnpackedMsi(i), 
    	    		   decl->GetUnpackedLsi(i), &coord);
	    CSimpleType* stype = new(CNode::CurrentHeap()) 
			CSimpleType(eINTEGER, cRANGE(cINT32(31),cINT32(0)) );
    	    index[i] = new(CNode::CurrentHeap()) CVar( 
    	    		CSymbol::GenSymbol("___index"), 
    	    		&coord, stype, FALSE );
    	    index[i]->SetDeclStatementCreated();
    	    declStmts = cLINK( declStmts, cVAR_DECL(index[i],NULL) );
    	    module->Add( index[i] );
        }
        for( int i = 0; i < dimensions; ++i ) {
	    if( nonx[i] ) {
                indexes = cELINK( indexes, dim[i]->Clone() );
	    } else {
                indexes = cELINK( indexes, cVAR_REF(index[i]) );
	    }	
        }
	CNode* lhs = target->Clone();
	if( select ) {
	    MASSERT( select->GetOp() == eARRAY );
	    select->Arg<CNode*>(0) = lhs;
	    select->Arg<CNode*>(1) = cELINK(indexes,select->Arg<CNode*>(1));
               lhs = select;
	} else {
	    if( indexes ) {
                lhs = cARRAY( lhs, indexes );
	    }
	}
	if( op == eASSIGN ) {
	    t = cLINK(cASSIGN( Clone(delay), lhs, width ),t);
	} else {
	    t = cLINK(cNBASSIGN( Clone(delay), lhs, width ),t);
	}
	for( int i = 0; i < dimensions; ++i ) {
	    if( index[i] ) {
	        t = cFOR( 
		    cASSIGN( NULL, cVAR_REF(index[i]), lb[i] ),
		    cLE( cVAR_REF(index[i]), mb[i] ),
		    cASSIGN( NULL, cVAR_REF(index[i]),
			cADD(cVAR_REF(index[i]),cINT32(1)) 
		    ),
		    t 
		);
	    }
	}
	if( conditional ) {
	    CNode* cond = NULL;
            for( int i = 0; i < indexCount; ++i ) {
	        if( dim[i]->IsConstant() || 
	            (!xOutDim[i] && IsNonX(dim[i])) ) { 
		    continue;
		}
		cond = cond ?  cCAT( cond, dim[i]->Clone() ) : dim[i]->Clone();
            }
	    if( altSelect && hasBitSelect &&
	        !dim.back()->IsConstant() && 
	        (xOutDim.back() || !IsNonX(dim.back()) ) ) { 
		cond = cond ?  cCAT( cond, dim.back()->Clone() ) : 
				           dim.back()->Clone();
	    }
	    if( altSelect && psType != 0 &&
	        !msb->IsConstant() && 
	        (xOutPartSelect || !IsNonX(msb) ) ) { 
		cond = cond ?  cCAT( cond, msb->Clone() ) : 
				           msb->Clone();
	    }
 	    if( cond ) {
	        CVector zero(1);
	        zero.LoadBinary( "0" );
	        CVector one(1);
	        one.LoadBinary( "1" );
		CNode* alt = NULL;
		CNode* indexes1 = NULL;
		if( cond->IsWidthVolatile() || 
		    !cond->IsWidthEvaluateable() || cond->GetWidth() != 1 ) {
		    cond = cRXOR(cond);
	        }
		if( indexCount == dimensions ) {
                    for( int i = 0; i < indexCount; ++i ) {
                        indexes1 = cELINK( indexes1, dim[i]->Clone() );
                    }
		    CNode* lhs1 = target->Clone();
		    if( indexes1 ) {
	                lhs1 = cARRAY( lhs1, indexes1 );
		    }
	            if( altSelect ) {
	                altSelect->Arg<CNode*>(0) = lhs1;
	                lhs1 = altSelect;
	            }
	            if( op == eASSIGN ) {
	                alt = cASSIGN( Clone(delay), lhs1, altWidth );
	            } else {
	                alt = cNBASSIGN( Clone(delay), lhs1, altWidth );
	            }
 		} else {
		    alt = CreateXAssignment( module, block, op, delay, target,
		    	dim, hasBitSelect, psType, msb, lsb, FALSE, 
		    	xOutPartSelect, xOutDim, TRUE );
		}
	        t = cCASE( cond,
		  cLIST(
		    cCASEITEM( cELIST(cVECTOR(zero),cVECTOR(one)), alt ),
		    cCASEITEM( NULL, t ) ) );
	    }
	}
	if( declStmts ) {
	    CBlock* block = new(CNode::CurrentHeap()) CBlock( 
			CSymbol::GenSymbol("___block"), &coord );
	    module->Add( block );
	    block->SetParent( module );
	    t = cBLOCK_REF( block, cLINK( declStmts, t) );
	}
        return t;
    }
    MASSERT( FALSE );
    return t;
} 

/*******************************************************
	LHSHash
	- class for managing list of lhs's seen
********************************************************/
class LHSHash {
    map<unsigned,vector<LHS> > lhsSeen;
public:
    LHSHash() {}
    ~LHSHash() {}
    LHS* Find( LHS& lhs );
    LHS* Insert( LHS& lhs );
};
	
LHS* LHSHash::Find( LHS& lhs )
{
    LHS* match = NULL;
    vector<LHS>::iterator lhsPtr;
    map<unsigned,vector<LHS> >::iterator mapPtr;
    unsigned id = lhs.Hash();
    mapPtr = lhsSeen.find(id);
    if( mapPtr == lhsSeen.end() ) {
	return NULL;
    }	
    for( lhsPtr = mapPtr->second.begin(); lhsPtr != mapPtr->second.end(); ++lhsPtr ) {
	if( lhs.Compatible( &*lhsPtr ) ) {
            return &*lhsPtr;
	}
        return NULL;
    }
}
LHS* LHSHash::Insert( LHS& lhs )
{
    unsigned id = lhs.Hash();
    lhsSeen[id].push_back( lhs );
    return &(lhsSeen[id].back());
}
     
/*******************************************************
	CreateLHSAssignments
	- Create assignment statements setting X's to list of LHS
********************************************************/
static CNode* CreateLHSAssignments( CBlock* module, CBlock* block, vector<LHS>& lhsList )
{
    CNode* t = NULL;
    vector<LHS>::iterator ptr;
    set<CDecl*> seen;
    map<CDecl*,set<int> > decl2BitVector;
    map<CDecl*,NodeOp_t> decl2Op;
    map<CDecl*,CNode*> decl2Delay;
    vector<LHS> lhsKeep;
    CSymtab<CDecl>& symtab = module->GetSymtab();

    set<CDecl*> forceWhole;
    /*
     * check each LHS to see if its indices or selects are dependent
     * upon other LHS in the block or out of scope variabls. If so, 
     * be pessimistic and X the whole structure.
     */
    // ??? mdh - need to make this less pessimistic. 
    // 1. do a deeper analyse to eliminate false temporal dependencies.
    // 2. do a deeper analyse to eliminate false subrange dependencies.
    set<CDecl*> lhsSet;
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {
	CNode* target = ptr->ref;
	CDecl* decl = target->Arg<CDecl*>(0);
	lhsSet.insert( decl );	
    }
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {
	if( ptr->psType != 0 ) {
            list<CDecl*> RHD;
     	    ptr->msb->PreVisit1( _CollectRHD, &RHD );		
	    list<CDecl*>::iterator rptr;
	    for( rptr = RHD.begin(); rptr != RHD.end(); ++rptr ) {
	        if( lhsSet.find(*rptr) != lhsSet.end() ) {
	    	    ptr->xOutPartSelect = 1;
		    break;
	        }
	    } 
	}	

	for( int i = 0; i < ptr->dim.size(); ++i ) {
            list<CDecl*> RHD;
	    ptr->dim[i]->PreVisit1( _CollectRHD, &RHD );
	    list<CDecl*>::iterator rptr;
	    for( rptr = RHD.begin(); rptr != RHD.end(); ++rptr ) {
	        if( lhsSet.find(*rptr) != lhsSet.end() ) {
	    	    ptr->xOutDim[i] = 1;
		    break;
	        }
	    } 
	}
        if( ptr->psType != 0 && 
	           !ExpressionInScope( block->GetSymtab(), ptr->msb ) ) {
	    ptr->xOutPartSelect = 1;
        }
	for( int i = 0; i < ptr->dim.size(); ++i ) {
	    if( !ExpressionInScope( block->GetSymtab(), ptr->dim[i] ) ) {
		ptr->xOutDim[i] = 1;
	    }
	}
    }

    /*
     * Fill in covering (vector range) info for each entry
     */
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {
	CNode* target = ptr->ref;
	CDecl* decl = target->Arg<CDecl*>(0);
	int initial = 0;
        int isArray = decl->GetNumberOfUnpackedDimensions() > 0;

	/*
 	 * initialized default covering info
 	 */
	ptr->fullyCovered      = 0;
	ptr->coveringInfoValid = 0;
	ptr->drop              = 0;

	/*
 	 * non part and bit references are fully covered
 	 */
	if( !ptr->hasBitSelect && !ptr->msb && !ptr->lsb ) {
	    ptr->fullyCovered      = 1;
	    ptr->coveringInfoValid = 1;
	    continue;
	}	
	/*
 	 * add covering info for bit vectors with non-volatile selects
 	 */
	if( ptr->hasBitSelect && !ptr->dim.back()->IsVolatile() &&
	    ptr->dim.back()->IsEvaluateable() ) {
	    int lsb = ptr->dim.back()->EvalINT32();
	    ptr->coveredBitVector.insert( lsb );
	    ptr->coveringInfoValid = 1;
	    continue;
	}

	/*
 	 * non const bit selects in loops are fully covered.
 	 * Note: this is because a range analysis would require
 	 * the for loop to be unrolled and the include procedural code
 	 * to be executed.
 	 */
	if( ptr->hasBitSelect && !ptr->dim.back()->IsConstant() &&
	    ptr->xOutDim.back() ) {
	    ptr->fullyCovered      = 1;
	    ptr->coveringInfoValid = 1;
	}

	/*
 	 * add covering info for part selects
 	 */
	if( ptr->psType == 0 && ptr->msb && ptr->lsb &&
	    !ptr->lsb->IsVolatile() && !ptr->msb->IsVolatile() &&
	    ptr->lsb->IsEvaluateable() && ptr->msb->IsEvaluateable() ) {
	    int lsb = ptr->lsb->EvalINT32();
	    int msb = ptr->msb->EvalINT32();
	    if( msb < lsb ) {
		int tmp = lsb;
		lsb = msb;
		msb = tmp;
	    }
	    for( int i = lsb; i <= msb; ++i ) {
	       ptr->coveredBitVector.insert( i );
	    }
	    ptr->coveringInfoValid = 1;
	    continue;
	}

	/*
 	 * add covering info for range selects
 	 */
	if( ptr->psType != 0 &&
	    (!ptr->lsb->IsVolatile() && !ptr->msb->IsVolatile() &&
	     ptr->lsb->IsEvaluateable() && ptr->msb->IsEvaluateable() ) ) {
	    int lsb = ptr->lsb->EvalINT32();
	    int msb = ptr->msb->EvalINT32();
	    if( decl->WidthDirection() > 0 ) {
		if( ptr->psType > 0 ) {
		    int temp = msb+lsb-1;
		    lsb = msb;
		    msb = temp;
		} else {
		    lsb = msb-lsb+1;
		}
	    } else {
		if( ptr->psType > 0 ) {
		    lsb = msb+lsb-1;
		} else {
		    int temp = msb-lsb+1;
		    lsb = msb;
		    msb = temp;
		}
	    }
		
	    if( msb < lsb ) {
		int tmp = lsb;
		lsb = msb;
		msb = tmp;
	    }
	    for( int i = lsb; i <= msb; ++i ) {
	        ptr->coveredBitVector.insert( i );
	    }
	    ptr->coveringInfoValid = 1;
	    continue;
	}

	/*
 	 * non const range selects in loops are fully covered.
 	 * Note: this is because a range analysis would require
 	 * the for loop to be unrolled and the include procedural code
 	 * to be executed.
 	 */
	if( ptr->psType != 0 &&
	    !ptr->lsb->IsConstant() && ptr->xOutPartSelect ) {
	    ptr->fullyCovered      = 1;
	    ptr->coveringInfoValid = 1;
	}
    }


    /*
     * mark all duplicates that are subsets of 
     * other entries to be dropped.
     */
    vector<LHS>::reverse_iterator rptr;
    vector<LHS>::reverse_iterator rptr1;
    int index;
    for( index = 0, rptr = lhsList.rbegin(); rptr != lhsList.rend(); ++rptr, ++index ) {
	int index1;
        for( index1 = 0, rptr1 = lhsList.rbegin(); rptr1 != lhsList.rend(); ++rptr1, ++index1 ) {
	    /*
  	     * if this is a subset of a another entry drop it
  	     */
	    if( index != index1 && !rptr1->drop && rptr->IsSubsetOf( &*rptr1 ) ) {
		rptr->drop = 1;
		break;
	    }
	}
    }

    /*
     * process each lhs and build covering set of assignments.
     * Note the set will not be the minimal set, only a sufficient set.
     */
    LHSHash	lhsHash;
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {
	CNode* target = ptr->ref;
	CDecl* decl = target->Arg<CDecl*>(0);
        int isArray = decl->GetNumberOfUnpackedDimensions() > 0;

	
	/*
 	 * if a subset of another entry, drop it
 	 */
	if( ptr->drop ) {
	    continue;
	}

	/*
 	 * drop arrays if support is disabled
 	 */
	if( disableArray && isArray ) {
	    continue;
	}
	
	/*
 	 * drop out of scope assignments
 	 */
	if( !ExpressionInScope( block->GetSymtab(), target ) ) {
	    continue;
	}

	/*
         * Check to see if we've seen compatible a lhs before
         * If not allocate add this one to hash table and 
         * lhs list.
         */
	LHS* lhsPtr;
	lhsPtr = lhsHash.Find(*ptr);
	if( !lhsPtr ) {
	    lhsPtr = lhsHash.Insert(*ptr);
	    goto KEEP;
	}

	
	/*
  	 * if this is a subset of a previous entry,
  	 * drop it.
  	 */
	if( ptr->IsSubsetOf( lhsPtr ) ) {
            continue;	   
	}

	/*
 	 * merge range info with existing entry and add to list
 	 */
	lhsPtr->fullyCovered |= ptr->fullyCovered;
	lhsPtr->coveredBitVector.insert( ptr->coveredBitVector.begin(),
					 ptr->coveredBitVector.end() );

KEEP:
	lhsKeep.push_back( *ptr );
    }

    /*
     * Create an assignment statement for each LHS
     */
    for( ptr = lhsKeep.begin(); ptr != lhsKeep.end(); ++ptr ) {
	t = cLINK( CreateXAssignment( module, block, ptr->op, ptr->delay, 
		   ptr->ref, ptr->dim, 
	  	   ptr->hasBitSelect, ptr->psType,
	           ptr->msb, ptr->lsb, TRUE, 
		   ptr->xOutPartSelect, ptr->xOutDim, FALSE ), t );	
    }

    return t;
}

/*******************************************************
 	XPropClockEvent
	- add xprop clock instrumentation to clock event 
 *******************************************************/
static void XPropClockEvent( CNode* event )
{
    /*
     * Add hidden attribute to create conditional code around
     * event expression to handle multi-edge intrumentation.
     */
    CNode* eventAttr = event->GetAttributes();
    if( !eventAttr ) {
        eventAttr = cATTRIBUTE(new(CNode::CurrentHeap()) 
    				CAttr(event->GetCoord()));
        event->SetAttributes( eventAttr );
    }
    eventAttr->Arg<CAttr*>(0)->AddAttribute( edgeControl, 
    				cSTRING(edgeAttrName) );
}

/*******************************************************
 	XPropClockBody
	- add xprop clock instrumentation to clock body
 *******************************************************/
static void XPropClockBody( CBlock* module, CBlock* block, 
			   NodeOp_t clockEdge,
			   CNode*   clockExp,
		           CNode_sp<CNode*> insertionPoint,
			   int dropTrailingComment )
{
    CNode* body = insertionPoint;
    /*
     * create xprop assignments for state elements
     */
    vector<LHS> lhsList;	
    lhsTranslateState = eTRANSLATE_ON;
    if( body != NULL ) {
        body->PreVisit1( _CollectLHS, &lhsList );
    }
    body = XPropStatement( module, block, body );
    CNode* xStmt = CreateLHSAssignments( module, block, lhsList );

    /*
     * create clock intrumentation
     */
    if( insertionClkBeginComment ) {
        CNode* b = (!body || body->GetOp() != eBLOCK_REF) ?
    			cBLOCK_REF( NULL, body) : body;	
        body = cLINK( b, insertionClkBeginComment->Clone() );
    }
    if( !dropTrailingComment && insertionClkEndComment ) {
        xStmt = cLINK( cBLOCK_REF( NULL, xStmt ), 
    			insertionClkEndComment->Clone() );
    }
    if( insertionClkEndComment ) {
        body = cLINK( insertionClkEndComment->Clone(), body );
    }
    if( clockEdge == ePOSEDGE ) {
        body = cIF( clockExp, body, cIF( cCOM(clockExp->Clone()), NULL, xStmt ) );
    } else {
        body = cIF( cCOM(clockExp->Clone()), body, cIF( clockExp, NULL, xStmt ) );
    }
    if( insertionClkBeginComment ) {
        body = cLINK( insertionClkBeginComment->Clone(), body );
    }
    insertionPoint = body;

}

/*******************************************************
 	XPropSimpleStateElement
	- xprop simple state element (clock, but no async controls).
	  module is block pointer to containing module.
	  block is block pointer to containing block.
	  event is event expression for element.
 *******************************************************/
static int XPropSimpleStateElement( CBlock* module, CBlock* block, CNode* event )
{
    MASSERT( event->GetOp() == eEVENT );
    /* 
     *  instrument simple clocking constructs 
     *  1) always @(posedge expression)
     *  2) always @(negedge expression)
     */
    CNode*   condition = event->Arg<CNode*>(0);
    CNode*   body = event->Arg<CNode*>(1);
    MASSERT( condition && (condition->GetOp() == ePOSEDGE ||
	                   condition->GetOp() == eNEGEDGE) );

    CNode_sp<CNode*>   insertionPoint = event->Arg<CNode*>(1);
    CNode*   clockExp = condition->Arg<CNode*>(0);
    NodeOp_t clockEdge = condition->GetOp();


    /*
     * transform:
     *    always @(POSEDGE exp)
     *	stmts
     * to:
     *    always @(POSEDGE exp)
     * 	if( expr ) 
     * 	    stmt
     *	else 
     *	    xprop'd LHS
     *
     * transform:
     *    always @(NEGEDGE exp)
     *	stmts
     * to:
     *    always @(NEGEDGE exp)
     * 	if( ~expr ) 
     * 	    stmt
     *	else 
     *	    xprop'd LHS
     *
     */

    XPropClockEvent( event );
    XPropClockBody( module, block, clockEdge, clockExp, 
		    insertionPoint, FALSE );
}

/*******************************************************
 	XPropStateElement
	- xprop state element. 
	  module is block pointer to containing module.
	  block is block pointer to containing block.
	  event is event expression for element.
	  return TRUE if element was sucessfully intrumented.
 *******************************************************/

static int _CollectPosedge( CNode* n, void* arg ) 
{
    set<CNode*>* s = static_cast<set<CNode*>*>(arg);
    switch( n->GetOp() ) {
    case ePOSEDGE:
	s->insert(n->Arg<CNode*>(0));
	break;
    default:
        return TRUE;
    }
}

static int _CollectNegedge( CNode* n, void* arg ) 
{
    set<CNode*>* s = static_cast<set<CNode*>*>(arg);
    switch( n->GetOp() ) {
    case eNEGEDGE:
	s->insert(cCOM(n->Arg<CNode*>(0)));
	break;
    default:
        return TRUE;
    }
}

static int _CollectIfExp( CNode* n, void* arg ) 
{
    set<CNode*>* s = static_cast<set<CNode*>*>(arg);
    switch( n->GetOp() ) {
    case eIF:
	s->insert(n->Arg<CNode*>(0));
	return FALSE;
    default:
        return TRUE;
    }
}
static void Elements2Vector(CNode_sp<CNode*> n, vector<CNode_sp<CNode*> >& v)
{
    if( static_cast<CNode*>(n) == NULL ) {
	return;
    }
    switch( n->GetOp() ) {
    case eLIST:
	Elements2Vector(n->Arg<CNode*>(0),v);
	Elements2Vector(n->Arg<CNode*>(1),v);
	break;
    case ePRAGMA:
    case eVRQ:
    case eCOMMENT:
	break;
    case eBLOCK_REF:
	Elements2Vector(n->Arg<CNode*>(1),v);
	break;
	break;
    default:
	v.push_back(n);
	break;
    }
}

static int XPropStateElement( CBlock* module, CBlock* block, CNode* event )
{
    CNode*   body = event->Arg<CNode*>(1);
    CNode*   exp = NULL;
    NodeOp_t edge = ePOSEDGE;
    CNode*   condition = event->Arg<CNode*>(0);

    /*
     * create list of all state element specifications
     */
     vector<CNode_sp<CNode*> > elements;
     Elements2Vector(event->Arg<CNode*>(1), elements);

    /*
     * inspect event expression
     */
    set<CNode*> posedgeList;
    set<CNode*> negedgeList;
    condition->PreVisit1( _CollectPosedge, &posedgeList );
    condition->PreVisit1( _CollectNegedge, &negedgeList );

    /*
     * validate each state element
     */
    vector<CNode_sp<CNode*> >::iterator ePtr;
    for( ePtr = elements.begin(); ePtr != elements.end(); ++ePtr ) {
	CNode_sp<CNode*> insertionPoint = *ePtr;
        set<CNode*> bothEdges;
	bothEdges.insert( posedgeList.begin(), posedgeList.end() );
	bothEdges.insert( negedgeList.begin(), negedgeList.end() );
    	set<CNode*> ifExpList;
	while( static_cast<CNode*>(insertionPoint) != NULL && 
	       insertionPoint->GetOp() == eIF ) {
	    CNode* cond = insertionPoint->Arg<CNode*>(0);
	    ifExpList.insert( cond );
            set<CNode*>::iterator ptr;
	    int found = 0;
	    for( ptr = posedgeList.begin(); ptr != posedgeList.end(); ++ptr ) {
		if( Equivalent( *ptr, cond ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    for( ptr = negedgeList.begin(); ptr != negedgeList.end(); ++ptr ) {
		if( Equivalent( *ptr, cond ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    if( !found ) {
		break;
	    }
	    insertionPoint = insertionPoint->Arg<CNode*>(2);
	}
	if( bothEdges.size() != 1 ) {
	    return FALSE;
	}
    }

    /*
     * add instrumentation to clock event
     */
    XPropClockEvent( event );
	
    /*
     * instrument each state element
     */
    for( ePtr = elements.begin(); ePtr != elements.end(); ++ePtr ) {
	CNode_sp<CNode*> insertionPoint = *ePtr;
        set<CNode*> bothEdges;
	bothEdges.insert( posedgeList.begin(), posedgeList.end() );
	bothEdges.insert( negedgeList.begin(), negedgeList.end() );
    	set<CNode*> ifExpList;
	int count = 0;
	while( static_cast<CNode*>(insertionPoint) != NULL && 
	       insertionPoint->GetOp() == eIF ) {
	    CNode* cond = insertionPoint->Arg<CNode*>(0);
	    ifExpList.insert( cond );
            set<CNode*>::iterator ptr;
	    int found = 0;
	    for( ptr = posedgeList.begin(); ptr != posedgeList.end(); ++ptr ) {
		if( Equivalent( *ptr, cond ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    for( ptr = negedgeList.begin(); ptr != negedgeList.end(); ++ptr ) {
		if( Equivalent( *ptr, cond ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    if( !found ) {
		break;
	    }

    	    /*
      	     * create xprop assignments for sync controls
             */
    	    vector<LHS> lhsList;	
            lhsTranslateState = eTRANSLATE_ON;
	    CNode* tStmt = insertionPoint->Arg<CNode*>(1);
	    CNode* fStmt = insertionPoint->Arg<CNode*>(2);
            if( tStmt != NULL ) {
        	tStmt->PreVisit1( _CollectLHS, &lhsList );
            }	     
            if( fStmt != NULL ) {
        	fStmt->PreVisit1( _CollectLHS, &lhsList );
            }	     
            tStmt = XPropStatement( module, block, tStmt );
    	    CNode* xStmt = CreateLHSAssignments( module, block, lhsList );

//             if( count == 0 && insertionClkEndComment ) {
//                 xStmt = cLINK( cBLOCK_REF( NULL, xStmt ), 
//       			        insertionClkEndComment->Clone() );
//             }
             if( insertionClkBeginComment || insertionClkEndComment ) {
                 CNode* fStmt = (!fStmt || fStmt->GetOp() != eBLOCK_REF) ?
       			    cBLOCK_REF( NULL, fStmt) : fStmt;	
	     }
//             if( insertionClkBeginComment ) {
//                 fStmt = cLINK( fStmt,insertionClkBeginComment->Clone() );
//             }
             if( insertionClkEndComment ) {
                 fStmt = cLINK( insertionClkEndComment->Clone(), fStmt );
             }
             CNode* n1 = cIF( cCOM(cond->Clone()), fStmt, xStmt );
	     CNode_sp<CNode*> newInsertionPointer =
			 newInsertionPointer = n1->Arg<CNode*>(1);
	     if( insertionClkEndComment ) {
	         newInsertionPointer = newInsertionPointer->Arg<CNode*>(1);
	     }
//	     if( insertionClkBeginComment ) {
//	         newInsertionPointer = newInsertionPointer->Arg<CNode*>(0);
//	     }
//             if( insertionClkBeginComment ) {
//                 n1 = cLINK( n1, insertionClkBeginComment->Clone() );
//             }
             if( insertionClkBeginComment || (count==0 && insertionClkEndComment) ) {
                 n1 = cBLOCK_REF( NULL, n1 );
	     }
             if( insertionClkBeginComment ) {
                 n1 = cLINK( insertionClkBeginComment->Clone(), n1 );
             }
             if( count == 0 && insertionClkEndComment ) {
                 n1 = cLINK( n1, insertionClkEndComment->Clone() );
             }
             CNode* body = cIF( cond, tStmt,  n1 );
            insertionPoint = body;
	    insertionPoint = newInsertionPointer;
	    count++;
	}
	CNode*   clockExp = *bothEdges.begin();
	/*
 	 * edge is always specified as POSEDGE as clockExp is already inverted
 	 */
        XPropClockBody( module, block, ePOSEDGE, clockExp, 
			insertionPoint, TRUE );
    }

    return TRUE;
}

/*******************************************************
	XPropStatement
	- add x propagation logic 
********************************************************/

static CNode* XPropStatement( CBlock* module, CBlock* block, CNode* n ) 
{
	if( n == NULL ) {
		return n;
	}

        if( IsNoxNode( n ) ) {
	    return n;
        }

	switch( n->GetOp() ) {
	case eDISABLE:
	case ePASSIGN:
	case eDEASSIGN:
	case eTRIGGER:
	case ePORT_DECL:
	case eNET_DECL:
	case eDEFPARAM:
	case ePARAM_DECL:
	case eVAR_DECL:
	case eGENVAR_DECL:
	case eFUNCTION_CALL:
	case eTASK_ENABLE:
	case eSYSTASK_CALL:
	case eINSTANCE_REF:
	case eGATE_REF:
	case ePORTLIST_END:
		break;
	case ePRAGMA:
		if( IsTranslateOn(n) ) {
			translateState = eTRANSLATE_ON;
		} else if( IsTranslateOff(n) ) {
			if( translateState == eTRANSLATE_OFF ) {
				message( n->GetCoord(), mNSTOS );
			}
			translateState = eTRANSLATE_OFF;
		}
		break;
	case eCOMMENT:
	case eVRQ:
		break;
	case eLIST:
		{
		  CNode* c = n;
		  while( 1 ) {
		      c->Arg<CNode*>(0) = 
			   XPropStatement( module, block, c->Arg<CNode*>(0) );
		      if( !c->Arg<CNode*>(1) || 
			   c->Arg<CNode*>(1)->GetOp() != eLIST ) {
		  	  c->Arg<CNode*>(1) = 
			      XPropStatement( module, block, c->Arg<CNode*>(1) );
			  break;
		      }
		      c = c->Arg<CNode*>(1);
		  }
	   	}
		break;	
	case eINIT:
		n->Arg<CNode*>(0) = XPropStatement( module, block, n->Arg<CNode*>(0) );
		break;
	case eEVENT:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;
	case eWHILE:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;	
	case eREPEAT:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;	
	case eFOREVER:
		n->Arg<CNode*>(0) = XPropStatement( module, block, n->Arg<CNode*>(0) );
		break;	
	case eGFOR:
	case eFOR:
		n->Arg<CNode*>(3) = XPropStatement( module, block, n->Arg<CNode*>(3) );
		break;	
	case eBLOCK_REF:
		{
		    CBlock* nameblock = n->Arg<CBlock*>(0);
		    if( nameblock ) {
			n->Arg<CNode*>(1) = XPropStatement( module, 
						nameblock, n->Arg<CNode*>(1) );
		    } else {
			n->Arg<CNode*>(1) = XPropStatement( module, 
						block, n->Arg<CNode*>(1) );
		    }
		}
		break;
	case eSPECIFY_REF:
		break;
	case eDELAY:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;	
	case eFUNCTION_DEF:
		{
			CBlock* block = n->Arg<CBlock*>(0);
			prependToModule = NULL;
			CNode* newBlock = XPropStatement( module, block, 
						block->GetCodeList() ); 
			block->SetCodeList( cLINK( prependToModule, newBlock) );	
		}
		break;
	case eMODULE_DEF:
		{
			CBlock* block = n->Arg<CBlock*>(0);
			prependToModule = NULL;
			CNode* newBlock = XPropStatement( block, block, 
						block->GetCodeList() ); 
			block->SetCodeList( cLINK( prependToModule, newBlock) );	
		}
		break;
	case eWAIT:
	case eCASSIGN:
	case eFORCE:
	case eRELEASE:
	case eTABLE:
		break;
	case eATTRIBUTE:
		break;
	case eGCASE:
		XPropCaseItem( module, block, n->Arg<CNode*>(1) );
		break;
	case eCASEX:
		message( n->GetCoord(), mCCBMS );
		XPropCaseItem( module, block, n->Arg<CNode*>(1) );
		break;
	case eGIF:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		n->Arg<CNode*>(2) = XPropStatement( module, block, n->Arg<CNode*>(2) );
		break;
/********************************************************
	these are the case that need special processing
 ********************************************************/

	case eASSIGN:
	case eNBASSIGN:
		if( translateState == eTRANSLATE_OFF || 
		     n->HasAttribute( noxAttribute ) ) {
			break;
		}
		{
		  CNode* rl = n->Arg<CNode*>(1);
		  vector<Pair> indexList;
		  rl->PreVisit1( _CollectIndex, (void*)&indexList );
	          if( indexList.size() == 0 ) {
			break;
		  }
		  vector<Pair>::iterator ptr;
		  CVector one(1);
		  one.LoadBinary( "1" );
		  CVector zero(1);
		  zero.LoadBinary( "0" );
		  CNode* cond = NULL;
		  CNode* t    = NULL;
		  set<CDecl*> seen;
		  map<CDecl*,NodeOp_t> decl2Op;
		  map<CDecl*,CNode*> decl2Delay;
		  for( ptr = indexList.begin(); 
					ptr != indexList.end(); ++ptr ) {
		      vector<CNode*>::iterator ptr1;
		      for( ptr1 = ptr->dim.begin(); 
					ptr1 != ptr->dim.end(); ++ptr1 ) {
			CNode* index = *ptr1;
			CNode* target = ptr->target;
	    		if( index && IsNonX(index) ) {
			    index = NULL;
			}
			if( index ) {
			    if( cond == NULL ) {
				cond = index->Clone();
			    } else {
				cond = cCAT( cond, index->Clone() );
			    }
			}
		      }
		      if( ptr->psType != 0 && ptr->msb ) { 
			CNode* index = ptr->msb;
			CNode* target = ptr->target;
	    		if( index && IsNonX(index) ) {
			    index = NULL;
			}
			if( index ) {
			    if( cond == NULL ) {
				cond = index->Clone();
			    } else {
				cond = cCAT( cond, index->Clone() );
			    }
			}
		      }
		   }
		   if( cond ) {
		       if( cond->IsWidthVolatile() || 
		           !cond->IsWidthEvaluateable() || cond->GetWidth() != 1 ) {
		           cond = cRXOR(cond);
		      }
		      for( ptr = indexList.begin(); ptr != indexList.end(); ++ptr ) {
			CNode* target = ptr->target;
			if( seen.find( target->Arg<CDecl*>(0) ) == seen.end() ) {
			    if( !disableArray || !ptr->dim.size() || 
				(ptr->hasBitSelect && ptr->dim.size() ==1) ) {
			 	vector<int> noXDim(ptr->dim.size(),0);
			        CNode* s = CreateXAssignment( module, 
					  block, n->GetOp(), 
					  NULL, target, ptr->dim, 
					  ptr->hasBitSelect, ptr->psType, 
					  ptr->msb, ptr->lsb, 
					  FALSE, FALSE, 
					  noXDim,
					  FALSE );	
			        t = cLINK( s, t );
			    }
			}
			seen.insert( target->Arg<CDecl*>(0) );
		     }
		     CNode* def = cCASEITEM( NULL, t );
		     if( insertionBeginComment ) {
		         def = cLINK( insertionBeginComment->Clone(), def );
		     }
		     if( insertionEndComment ) {
		         def = cLINK( def, insertionEndComment->Clone() );
		     }
		     n = cCASE( cond, cLIST( 
			    cCASEITEM( cELIST(cVECTOR(zero),cVECTOR(one)), n ),
			    def 
			)
		      );
		  }
	       }		  
	       break;
	case eCASEZ:
	case eCASE:
		if( translateState == eTRANSLATE_OFF || disableCase ||
		    n->HasAttribute( noxAttribute ) ) {
			n->Arg<CNode*>(1) = XPropCaseItem( module, block, n->Arg<CNode*>(1) );
			break;
		}
		{

			CNode* r = n->Arg<CNode*>(1);
			CNode* l = n->Arg<CNode*>(0);
			vector<LHS> lhsList;	
			lhsTranslateState = translateState;
			if( r != NULL ) {
				r->PreVisit1( _CollectLHS, &lhsList );
			}
			r = XPropCaseItem( module, block, r );
			n->Arg<CNode*>(1) = r;
			int instrumentExpression = !IsNonX(l);
			if( lhsList.size() == 0 ) {
				break;
			}
			list<CNode*> caseList;
			r->PreVisit1( _CollectCases, &caseList );

			if( caseList.size() == 0 && !instrumentExpression ) {
			    break;
			}

			/*
			 * find default case item
			 */
		        CNode* deflt = NULL;
		        n->Arg<CNode*>(1)->PreVisit1(_CollectCaseDefault,&deflt);


			CNode* innerCond = NULL;
			CNode* outerCond = NULL;
		        CVector one(1);
		        one.LoadBinary( "1" );
		        CVector zero(1);
		        zero.LoadBinary( "0" );
			if( instrumentExpression ) {
			    innerCond = l->Clone();
			}
			if( caseList.size() != 0 ) {
			    list<CNode*>::iterator ptr;  
			    for( ptr = caseList.begin(); ptr != caseList.end(); ++ptr ) {
				CNode* next = EList2Cat( *ptr );
				outerCond = outerCond ? 
						cCAT( outerCond, next ) : next;
			    }
			}
			if( innerCond ) {
			    if( innerCond->IsWidthVolatile() || 
				!innerCond->IsWidthEvaluateable() ||
		                innerCond->GetWidth() != 1 ) {
			        innerCond = cRXOR(innerCond);
			    }
			}
			if( outerCond ) {
			    if( outerCond->IsWidthVolatile() || 
				!outerCond->IsWidthEvaluateable() ||
			        outerCond->GetWidth() != 1 ) {
			        outerCond = cRXOR(outerCond);
			    }
			}


			if( innerCond ) {
		            if( !deflt ) {
			        deflt = cCASEITEM( NULL, NULL, NULL );
		     	        CNode* deflt1 = deflt;
				if( insertionBeginComment ) {
				    deflt1 = cLINK( insertionBeginComment->Clone(), deflt1 );
				}
				if( insertionEndComment ) {
		     	            deflt1 = cLINK( deflt1, insertionEndComment->Clone() );
				}
		                n->Arg<CNode*>(1) = cLIST( n->Arg<CNode*>(1), deflt1 );
			    }
			    CNode* t = CreateLHSAssignments( module, block, lhsList );
			    CNode* def = cCASEITEM( NULL, t );
			    if( deflt->Arg<CNode*>(1) && insertionBeginComment ) {
		     	        def = cLINK( insertionBeginComment->Clone(), def );
			    }
			    if( deflt->Arg<CNode*>(1) && insertionEndComment ) {
		     	        def = cLINK( def, insertionEndComment->Clone() );
			    }
		            CNode* newDeflt = cCASE( innerCond,
			        cLIST( 
			            cCASEITEM( 
			    	    cELIST(cVECTOR(zero),cVECTOR(one)), 
			    	    deflt->Arg<CNode*>(1) ),
			            def ) );
			    deflt->Arg<CNode*>(1) = newDeflt;
			}
			if( outerCond ) {
			    CNode* t = CreateLHSAssignments( module, block, lhsList );
			    CNode* def = cCASEITEM( NULL, t );
			    if( insertionBeginComment ) {
		     	        def = cLINK( insertionBeginComment->Clone(), def );
			    }
			    if( insertionEndComment ) {
		     	        def =cLINK( def, insertionEndComment->Clone() );
			    }
		            CNode* newN = cCASE( outerCond,
			        cLIST( 
			            cCASEITEM( 
			    	    cELIST(cVECTOR(zero),cVECTOR(one)), 
			    	    n ),
			            def ) );
			    newN->SetAttributes( n->GetAttributes() );
			    n = newN;
			}
		}
		break;
	case eIF:
		if( translateState == eTRANSLATE_OFF || disableIf ||
		    n->HasAttribute( noxAttribute ) ) {
			n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
			n->Arg<CNode*>(2) = XPropStatement( module, block, n->Arg<CNode*>(2) );
			break;
		}
		{
			CNode* condition  = n->Arg<CNode*>(0);
			CNode* tStmt = n->Arg<CNode*>(1);
			CNode* fStmt = n->Arg<CNode*>(2);
			vector<LHS> lhsList;	
			lhsTranslateState = translateState;
			if( tStmt != NULL ) {
				tStmt->PreVisit1( _CollectLHS, &lhsList );
			}
			if( fStmt != NULL ) {
				fStmt->PreVisit1( _CollectLHS, &lhsList );
			}
			tStmt = XPropStatement( module, block, tStmt );
			fStmt = XPropStatement( module, block, fStmt );
			n->Arg<CNode*>(1) = tStmt;
			n->Arg<CNode*>(2) = fStmt;
			if( IsNonX(condition) ) {
				break;
			}
			if( lhsList.size() == 0 ) {
				break;
			}
			CNode* xStmt = CreateLHSAssignments( module, block, lhsList );
//#define OLD 1
#ifdef OLD
			if( insertionBeginComment ) {
			    CNode* b = (!fStmt || fStmt->GetOp() != eBLOCK_REF) ?
					cBLOCK_REF( NULL, fStmt) : fStmt;	
			    fStmt = cLINK( b, insertionBeginComment->Clone() );
			}
			if( insertionEndComment ) {
			    xStmt = cLINK( cBLOCK_REF( NULL, xStmt ), 
						insertionEndComment->Clone() );
			}
#else
			if( !fStmt && insertionBeginComment ) {
			    CNode* b = (!tStmt || tStmt->GetOp() != eBLOCK_REF) ?
					cBLOCK_REF( NULL, tStmt) : tStmt;	
			    tStmt = cLINK( b, insertionBeginComment->Clone() );
			}
			if( fStmt && 
			    (insertionBeginComment || insertionEndComment) ) {
			    CNode* b = (fStmt->GetOp() != eBLOCK_REF) ?
					cBLOCK_REF( NULL, fStmt) : fStmt;	
			    b = cLINK( insertionEndComment->Clone(), b );
			    fStmt = cLINK(b, insertionBeginComment->Clone());
			}
			if( insertionEndComment ) {
			    xStmt = cLINK( cBLOCK_REF( NULL, xStmt ), 
						insertionEndComment->Clone() );
			}
#endif
			CNode* n1 = cIF( cCOM(condition), fStmt, xStmt );
			if( fStmt && insertionBeginComment ) {
			     n1 = cLINK( insertionBeginComment->Clone(), n1 );
			}
			CNode* n_new = cIF( condition, tStmt, n1 );
			n_new->SetAttributes( n->GetAttributes() );
			n = n_new;
		}
		break;	
	case eALWAYS: {
		int disabled = translateState == eTRANSLATE_OFF || !enableClock;
		int unsupported = FALSE;
		/* 
		 * check to see if always statement is a state element 
		 */
		CNode* event = n->Arg<CNode*>(0);
		int    edgeCount = 0;
		if( event && event->GetOp() == eEVENT ) {
		    if( event->HasAttribute( noxAttribute ) ) {
			disabled = 1;
		    }
    		    if( event->Arg<CNode*>(0) && 
			IsNonX(event->Arg<CNode*>(0)) ) {
			disabled = 1;
		    }
		    CNode* condition = event->Arg<CNode*>(0);
		    if( condition ) {
	                condition->PostVisit1( _CountEdges, &edgeCount );
		    }
		}	
		/*
 		 * handle simple sync clock blocks as special case
 		 */
		if( !disabled && edgeCount == 1 ) {
			XPropSimpleStateElement( module, block, event );
		/*
 		 * handle complex state element case
 		 */
		} else if( !disabled && edgeCount > 1 ) {
		    unsupported = !XPropStateElement( module, block, event );
		    if( unsupported ) {
        	        message( event->GetCoord(), mUESCI );
		    }
		}
		/*
 		 * handle all combinational, unsupported state elements,
 		 * and blocks with clock instrumentation disabled
 		 */
		if( disabled || unsupported || edgeCount == 0 ) {
		    n->Arg<CNode*>(0) = XPropStatement( module, 
					block, n->Arg<CNode*>(0) );
		}
		} break;
	default:
		MASSERT( FALSE );
	}

	return n;
	
}

/************************************************
	XPropTernary
	- Instrument ? operator
	  s ? a : b => out = s ? a : (~s ? b : n'bx)
 ************************************************/
int XPropTernary( CNode* n, void* )
{
    MASSERT( n );
    if( IsNoxNode( n ) ) {
	return 0;
    }
    switch( n->GetOp() ) {
    case eHOOK: 
	{
	   CNode* fexp = n->Arg<CNode*>(2);
	   CNode* texp = n->Arg<CNode*>(1);
	   CNode* xexp = NULL;
	   CNode* cond = n->Arg<CNode*>(0);
	   if( n->HasAttribute( noxAttribute ) ) {
		break;
	   }
	   /*
 	    * don't instrument if condition cannot be X
 	    */
	   if( IsNonX( cond ) ) {
		break;
	   }
	   NodeType_t ttype = texp->GetNodeType();
	   if( ttype != eS && ttype != eB ) {
	       break;
	   }
	   NodeType_t ftype = fexp->GetNodeType();
	   if( ftype != eS && ftype != eB ) {
		break;
	   }
	   if( fexp->IsWidthVolatile() || !fexp->IsWidthEvaluateable() ) {
		CNode* w = fexp->GetWidthExp();
		if( !w ) {
		    message( n->GetCoord(), mCDWOO );
		} else {
	           CVector x(1);
	           x.LoadBinary( "x" );
	           x.SetPreferredBase(2);
		   xexp = cREP(w,cVECTOR(x));
		}
	   }
	   if( !xexp ) {
	       CVector x(fexp->GetWidth());
	       x.SetToX();
	       x.SetPreferredBase( 16 );
	       xexp = cVECTOR(x);
	   }
	   n->Arg<CNode*>(0)->PreVisit1( XPropTernary, NULL );
	   n->Arg<CNode*>(1)->PreVisit1( XPropTernary, NULL );
	   n->Arg<CNode*>(2)->PreVisit1( XPropTernary, NULL );
	   CNode* newHook = cHOOK( cCOM(n->Arg<CNode*>(0)->Clone()),
				   n->Arg<CNode*>(2), xexp );
	   n->Arg<CNode*>(2) = newHook;
	}
	return 0;
    default:;
    }
    return 1;
}

/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CXProp::Process( list<CElement>& inputList,
			list<CElement>& outputList ) 
{ 

    /* 
     * allow synth on/off comments to be overridden
     */
    list<string> commentList1;
    GetPlusArgs( "xprop-begin=", commentList1 );
    insertionBeginComment = StringList2Comments( commentList1 );

    list<string> commentList2;
    GetPlusArgs( "xprop-end=", commentList2 );
    insertionEndComment = StringList2Comments( commentList2 );

    list<string> commentList3;
    GetPlusArgs( "xprop-clk-begin=", commentList3 );
    insertionClkBeginComment = StringList2Comments( commentList3 );

    list<string> commentList4;
    GetPlusArgs( "xprop-clk-end=", commentList4 );
    insertionClkEndComment = StringList2Comments( commentList4 );

    disableIf       = GetPlusArg( "xprop-disable-if" ) != NULL;
    disableCase     = GetPlusArg( "xprop-disable-case" ) != NULL;
    disableSelect   = GetPlusArg( "xprop-disable-select" ) != NULL;
    disableArray    = GetPlusArg( "xprop-disable-array" ) != NULL;
    disableTernary  = GetPlusArg( "xprop-disable-ternary" ) != NULL;
    integerIsNoX    = GetPlusArg( "xprop-int-nox" ) != NULL;
    noxAttribute    = GetPlusArg( "xprop-nox-attr" );
    if( noxAttribute ) {
	// skip over "="
	noxAttribute++;
    }
    enableClock  = GetPlusArg( "xprop-enable-clock" ) != NULL;
    edgeControl	 = GetPlusArg( "xprop-clk-edge-control" );
    if( edgeControl ) {
	if( edgeControl[1] == '!' ) {
           edgeAttrName = strdup( &edgeControl[2] );
           edgeControl = ".dump_strip_edges_ifndef";
	} else {
           edgeAttrName = strdup( &edgeControl[1] );
           edgeControl = ".dump_strip_edges_ifdef";
	}
    } else {
       edgeControl = ".dump_strip_edges_ifdef";
       edgeAttrName = "XPROP_BOTH_EDGES";
    }



    list<CElement>::iterator ptr;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
	/*
	 * scan though code list adding
	 * code to propagate x's
	 */
	CNode* code = ptr->Code();
	if( code && !disableTernary ) {
	    /*
 	     * Accellerate label operations to speed up
 	     * processing. Node label info should not change
 	     * due to this processing stage.
 	     */
	    CNode::EnableLabelCache();
	    code->PreVisit1( XPropTernary, NULL);
	    CNode::DisableAndClearLabelCache();
	}
	CNode::EnableLabelCache();
	code = XPropStatement( NULL, NULL, code );
	CNode::DisableAndClearLabelCache();
	outputList.push_back( CElement( ptr->Filename(), ptr->Filename()!=NULL, code ) );
    }
}

/**
 * \endcond
 */
